function demo1();

% This demo load 16 random VOC 2012 validation segmentation posteriors,
% subsampled in spatial size.
% It then evaluates the proposed method, the greedy method, and the
% inverse-weighted baseline method and visualizes the optimal decisions.
%
% If you use this code, please cite the original paper, thanks:
%
% @InProceedings{ nowozin2014iou,
%     title = "Optimal Decisions from Probabilistic Models:
%         the Intersection-over-Union Case",
%     author = "Sebastian Nowozin",
%     booktitle = "{IEEE} Conference on Computer Vision and Pattern Recognition (CVPR)",
%     publisher = "IEEE Computer Society",
%     year = "2014",
% }

clear;
load voc2012val_demo.mat;	% load posteriors
load voc_cmap.mat;		% load familiar VOC segmentation colormap

options=[];
options.init='random';	% random initialization

% 1. Use LFP local search algorithm (proposed in the paper)
options.algorithm='lfp';
options.sweep_count=5;		% number of outer sweep iterations
options.subset_size=2000;	% number of variables to process at a time
[Ylfp,rec,obj_lfp]=voc_optimize(P,options);

% 2. Use greedy local search algorithm as proposed by Tarlow and Adams
% (This is fast and competetive and a practical choice.)
options=[];
options.algorithm='greedy';
[Ygreedy,~,obj_greedy]=voc_optimize(P,options);

% 3. Use inverse-weighted baseline (iwMAP)
options_iw=[];
options_iw.init='inverseweighted';
options_iw.sweep_count=0;
[Yiw,~,obj_iw]=voc_optimize(P,options_iw);

% Plot lfp objective over iterations
figure;
plot(1:size(rec,2), rec(1,:), 'k-x');
xlabel('sweep iteration');
ylabel('LFP, believed IoU');
title('Example: 16 PASCAL VOC 2012 val images');

% Plot segmentations
draw_y(Ylfp, voc_cmap, sprintf('Proposed method, believed IoU %f',mean(obj_lfp)));
draw_y(Ygreedy, voc_cmap, sprintf('Greedy method, believed IoU %f',mean(obj_greedy)));
draw_y(Yiw,voc_cmap, sprintf('inverse-weighted method, believed IoU %f',mean(obj_iw)));


function draw_y(Y,voc_cmap,titlestr);
figure;
for i=1:numel(Y)
	subplot(4,4,i);
	imagesc(Y{i}); axis image; axis off;
	colormap(voc_cmap);
end

set(gcf,'NextPlot','add');
axes;
h = title(titlestr);
set(gca,'Visible','off');
set(h,'Visible','on'); 

