
#include <algorithm>
#include <iostream>
#include <cassert>
#include <map>
#include <cmath>

#include <mex.h>

#include "matlab_helpers.h"
#include "matlab_out.h"

matlab_out mp;
std::streambuf* matlab_out::mp_backup = 0;

void MatlabCPPInitialize(bool verbose) {
	matlab_out::OutInitialize(&mp, verbose);
}

void MatlabCPPExit() {
	matlab_out::OutExit();
}

bool HasOption(const mxArray* opt_s, const std::string& opt_name) {
	if (opt_s == 0)
		return (false);

	if (mxIsStruct(opt_s) == false) {
		mexErrMsgTxt("Passed options is not a structure.\n");
		assert(0);
		return (false);
	}
	int fn = mxGetFieldNumber(opt_s, opt_name.c_str());
	return (fn >= 0);
}

const mxArray* GetOption(const mxArray* opt_s, const std::string& opt_name) {
	assert(HasOption(opt_s, opt_name));
	int fn = mxGetFieldNumber(opt_s, opt_name.c_str());
	assert(fn >= 0);

	const mxArray* res = mxGetFieldByNumber(opt_s, 0, fn);
	assert(res != 0);
	return (res);
}

double GetScalarOption(const mxArray* opt_s,
	const std::string& opt_name) {
	const mxArray* opt = GetOption(opt_s, opt_name);
	if (mxIsNumeric(opt) == false || mxGetM(opt) != 1 || mxGetN(opt) != 1) {
		mexErrMsgTxt("Invalid use of a scalar-valued option.\n");
		return (0);
	}
	return (mxGetScalar(opt));
}

double GetScalarDefaultOption(const mxArray* opt_s,
	const std::string& opt_name, double default_value) {
	if (HasOption(opt_s, opt_name) == false)
		return (default_value);

	return (GetScalarOption(opt_s, opt_name));
}

std::string GetStringDefaultOption(const mxArray* opt_s,
	const std::string& opt_name, const std::string& default_value) {
	if (HasOption(opt_s, opt_name) == false)
		return (default_value);

	return (GetMatlabString(GetOption(opt_s, opt_name)));
}

unsigned int GetIntegerDefaultOption(const mxArray* opt_s,
	const std::string& opt_name, double default_value) {
	return (static_cast<unsigned int>(GetScalarDefaultOption(opt_s,
		opt_name, default_value)));
}

std::string GetMatlabString(const mxArray* m_str) {
	assert(mxIsChar(m_str));
	int char_count = static_cast<int>(mxGetNumberOfElements(m_str) + 1);
	char* buf = static_cast<char*>(mxCalloc(char_count, sizeof(char)));
	mxGetString(m_str, buf, char_count);

	std::string res(buf);
	mxFree(buf);

	return (res);
}

void GetMatlabVector(const mxArray* vin, std::vector<unsigned int>& vout) {
	if (mxIsDouble(vin) == false) {
		mexErrMsgTxt("Passed array is no double array.\n");
		assert(0);
		return;
	}
	size_t vec_count = mxGetNumberOfElements(vin);
	vout.resize(vec_count);
	const double* vec_p = mxGetPr(vin);
	for (unsigned int n = 0; n < vec_count; ++n)
		vout[n] = static_cast<unsigned int>(vec_p[n]);
}

void GetMatlabVector(const mxArray* vin, std::vector<double>& vout) {
	if (mxIsDouble(vin) == false) {
		mexErrMsgTxt("Passed array is no double array.\n");
		assert(0);
	}
	size_t vec_count = mxGetNumberOfElements(vin);
	vout.resize(vec_count);
	const double* vec_p = mxGetPr(vin);
	for (unsigned int n = 0; n < vec_count; ++n)
		vout[n] = vec_p[n];
}

void GetMatlabVector(const mxArray* vin, std::vector<float>& vout) {
	if (mxIsSingle(vin) == false) {
		mexErrMsgTxt("Passed array is no single array.\n");
		assert(0);
	}
	size_t vec_count = mxGetNumberOfElements(vin);
	vout.resize(vec_count);
	const float* vec_p = static_cast<const float*>(mxGetData(vin));
	for (unsigned int n = 0; n < vec_count; ++n)
		vout[n] = vec_p[n];
}

void GetMatlabVector(const mxArray* f_data,
	std::vector<double>& data_elem,
	std::vector<unsigned int>& data_idx) {
	assert(mxIsSparse(f_data) == true);
	assert(mxGetN(f_data) == 1);
	mwSize nzmax = mxGetNzmax(f_data);
	data_elem.resize(nzmax);
	data_idx.resize(nzmax);

	unsigned int nnz = 0;
	mwIndex* ir = mxGetIr(f_data);
	mwIndex* jc = mxGetJc(f_data);
	double* f_data_p = mxGetPr(f_data);

	for (size_t j = jc[0]; j < jc[1]; ++j, ++nnz) {
		data_elem[nnz] = f_data_p[j];
		data_idx[nnz] = static_cast<unsigned int>(ir[j]);
	}

	data_elem.resize(nnz);
	data_idx.resize(nnz);
}

