
#include <algorithm>
#include <numeric>
#include <map>
#include <utility>
#include <functional>
#include <cassert>
#include <cstdint>

#ifdef _OPENMP
#include <omp.h>
#endif

#include <mex.h>

#include "matlab_helpers.h"

// Project includes
#include "voc.h"
#include "timing.h"

// function [Y,rec,obj]=mex_voc_optimize(X,options);
void mexFunction(int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[]) {
	MatlabCPPInitialize(true);
	if (nlhs < 1 || nlhs > 3 || nrhs < 1 || nrhs > 2) {
		mexErrMsgTxt("Wrong number of arguments.\n");
		MatlabCPPExit();
		return;
	}

	// Get option values
	const mxArray* opt_s = 0;
	if (nrhs >= 2 && mxIsEmpty(prhs[1]) == false)
		opt_s = prhs[1];

	typedef voc::prediction_function<double> voc_t;
	typedef voc_t::class_t class_t;
	voc_t::pfunc_t P;
	size_t n = 0;
	size_t class_count = 0;

	std::map<size_t, size_t, std::greater<size_t>> i_to_image;
	std::vector<void *> image_to_p;
	std::vector<void *> image_to_y;

	const mxArray* Yinit = HasOption(opt_s, "Yinit") ?
		GetOption(opt_s, "Yinit") : 0;
	if (Yinit != 0 && HasOption(opt_s, "init")) {
		mexErrMsgTxt("When supplying options.Yinit, you cannot use options.init!\n");
		MatlabCPPExit();
		return;
	}
	voc_t::initfunc_t Yinit_func;

	// In case X is a (class_count,N) array
	bool is_array = false;
	if (mxIsSingle(prhs[0]) || mxIsDouble(prhs[0])) {
		is_array = true;
		class_count = mxGetM(prhs[0]);
		if (class_count < 2) {
			mexErrMsgTxt("At least two classes required.\n");
			MatlabCPPExit();
			return;
		}
		n = mxGetN(prhs[0]);
		if (n == 0) {
			mexErrMsgTxt("Need at least one instance.\n");
			MatlabCPPExit();
			return;
		}

		if (mxIsSingle(prhs[0])) {
			const float* X_pr = static_cast<const float*>(mxGetData(prhs[0]));
			P = [X_pr, class_count](size_t i, size_t c) -> double {
				return X_pr[i*class_count + c]; };
		} else {
			const double* X_pr = static_cast<const double*>(mxGetData(prhs[0]));
			P = [X_pr, class_count](size_t i, size_t c) -> double {
				return X_pr[i*class_count + c]; };
		}

		// Handle initialization
		if (Yinit != 0) {
			if (mxGetNumberOfElements(Yinit) != n
				|| mxIsClass(Yinit, "uint8") == false)
			{
				mexErrMsgTxt("options.Yinit needs to be uint8 array and "
					"match problem dimensions.\n");
				MatlabCPPExit();
				return;
			}
			std::uint8_t* Yinit_p = static_cast<std::uint8_t*>(mxGetData(Yinit));
			Yinit_func = [Yinit_p](size_t i) -> class_t { return Yinit_p[i]; };
		}
	} else if (mxIsCell(prhs[0])) {
		is_array = false;
		size_t M = mxGetNumberOfElements(prhs[0]);
		if (M == 0) {
			mexErrMsgTxt("Need at least one image posterior in cellarray.\n");
			MatlabCPPExit();
			return;
		}

		const mxArray** im = static_cast<const mxArray**>(mxGetData(prhs[0]));
		if (mxGetNumberOfDimensions(im[0]) != 3) {
			mexErrMsgTxt("When using cellarray argument each element needs to "
				"be three dimensional (class_count,h,w).\n");
			MatlabCPPExit();
			return;
		}
		const mwSize* dims = mxGetDimensions(im[0]);
		class_count = dims[0];
		bool is_single = mxIsSingle(im[0]);

		size_t pix_count = 0;
		for (size_t j = 0; j < M; ++j) {
			i_to_image.insert(std::pair<size_t, size_t>(pix_count, j));

			if (mxGetNumberOfDimensions(im[j]) != 3) {
				mexErrMsgTxt("When using cellarray argument each element needs to "
					"be three dimensional (class_count,h,w).\n");
				MatlabCPPExit();
				return;
			}

			dims = mxGetDimensions(im[j]);
			bool j_is_single = mxIsSingle(im[j]);
			if (j_is_single != is_single || dims[0] != class_count) {
				mexErrMsgTxt("All images must be of the same type (single or "
					"double) and have the same number of elements in the "
					"first dimension.\n");
				MatlabCPPExit();
				return;
			}
			image_to_p.push_back(mxGetData(im[j]));

			pix_count += dims[1] * dims[2];
		}
		n = pix_count;	// total number of pixels

		// Define probability access function
		P = [is_single, n, class_count, &i_to_image, &image_to_p](size_t i, size_t c)
			-> double
		{
			assert(i < n && c < class_count);

			auto imi = i_to_image.lower_bound(i);
			assert(imi != i_to_image.end());
			size_t j = imi->second;		// image index
			assert(i >= imi->first);
			size_t i_pix = i - imi->first;	// pixel within image

			assert(j < image_to_p.size());

			if (is_single) {
				const float* im = static_cast<const float*>(image_to_p[j]);
				return im[i_pix*class_count + c];
			} else {
				const double* im = static_cast<const double*>(image_to_p[j]);
				return im[i_pix*class_count + c];
			}
		};

		// Handle initialization
		if (Yinit != 0) {
			if (mxIsCell(Yinit) == false || mxGetNumberOfElements(Yinit) != M) {
				mexErrMsgTxt("options.Yinit needs to be a cell array "
					"if image array is used.\n");
				MatlabCPPExit();
				return;
			}

			const mxArray** init_a = static_cast<const mxArray**>(mxGetData(Yinit));
			size_t pix_count2 = 0;
			for (size_t j = 0; j < M; ++j) {
				if (mxIsClass(init_a[j], "uint8") == false) {
					mexErrMsgTxt("Elements in options.Yinit need to be uint8 arrays.\n");
					MatlabCPPExit();
					return;
				}
				image_to_y.push_back(mxGetData(init_a[j]));
				pix_count2 += mxGetNumberOfElements(init_a[j]);
			}
			if (pix_count2 != pix_count) {
				mexErrMsgTxt("Pixel count in options.Yinit does not match "
					"image pixel count.\n");
				MatlabCPPExit();
				return;
			}

			Yinit_func = [n, &i_to_image, &image_to_y](size_t i) -> class_t {
				assert(i < n);
				auto imi = i_to_image.lower_bound(i);
				assert(imi != i_to_image.end());
				size_t j = imi->second;		// image index
				assert(i >= imi->first);
				size_t i_pix = i - imi->first;	// pixel within image

				assert(j < image_to_p.size());
				const std::uint8_t* y_im =
					static_cast<const std::uint8_t*>(image_to_y[j]);
				return y_im[i_pix];
			};
		}
	}

	// Create VOC object and initialize
	voc_t pred(P, class_count, n);

	if (Yinit != 0) {
		pred.initialize(Yinit_func);
	} else {
		// Initialization type
		auto init_s = GetStringDefaultOption(opt_s, "init", "random");
		voc_t::init_type init = voc_t::random_assignment;
		if (init_s == "random") {
			init = voc_t::random_assignment;
		} else if (init_s == "map") {
			init = voc_t::map_assignment;
		} else if (init_s == "inverseweighted") {
			init = voc_t::inverseweighted_assignment;
		} else {
			mexErrMsgTxt("In options.init the choices are 'random', 'map', "
				"or 'inverseweighted'.\n");
			MatlabCPPExit();
			return;
		}
		pred.initialize(init);
	}

	auto algorithm_s = GetStringDefaultOption(opt_s, "algorithm", "lfp");
	if (algorithm_s != "lfp" && algorithm_s != "greedy") {
		mexErrMsgTxt("In options.algorithm the choices are 'lfp' or 'greedy'.\n");
		MatlabCPPExit();
		return;
	}

	// Other options
	unsigned int sweep_count = GetIntegerDefaultOption(opt_s, "sweep_count", 30);
	unsigned int subset_size = GetIntegerDefaultOption(opt_s, "subset_size", 0);

	// Recorded iteration statistics
	std::vector<double> record_obj;
	std::vector<double> record_time;

	timer total;
	total.start();
	bool change = true;
	for (size_t iter = 1; change && iter <= sweep_count; ++iter) {
		double obj = pred.compute_objective();
		std::cout << "[voc] iter " << iter << "  score " << obj << std::endl;

		// Record
		record_obj.push_back(obj);
		record_time.push_back(total.elapsed_s());

		timer t;
		t.start();
		if (algorithm_s == "lfp") {
			if (subset_size == 0) {
				pred.relabel_sweep();
			} else {
				pred.relabel_sweep_subset_pp(subset_size);
			}
		} else if (algorithm_s == "greedy") {
			change = (pred.greedy_sweep() > 0);
		} else {
			assert(0);
		}
		t.stop();
		std::cout << "[voc]    sweep took " << t.elapsed_s() << "s" << std::endl;
	}
	double final_obj = pred.compute_objective();
	std::cout << "[voc] final score " << final_obj << std::endl;

	// Copy solution
	auto& y_sol = pred.get_solution();
	assert(y_sol.size() == n);

	if (is_array) {
		plhs[0] = mxCreateNumericMatrix(1, n, mxUINT8_CLASS, mxREAL);
		std::uint8_t* Y = static_cast<std::uint8_t*>(mxGetData(plhs[0]));
		for (size_t i = 0; i < n; ++i)
			Y[i] = y_sol[i] + 1;
	} else {
		size_t M = mxGetNumberOfElements(prhs[0]);

		mwSize dim_0 = M;
		mxArray* Ycell = 0;
		Ycell = mxCreateCellArray(1, &dim_0);
		assert(Ycell != NULL);
		plhs[0] = Ycell;

		const mxArray** im = static_cast<const mxArray**>(mxGetData(prhs[0]));
		size_t pi = 0;
		for (size_t j = 0; j < M; ++j) {
			const mwSize* dims = mxGetDimensions(im[j]);

			mxArray* Yj = mxCreateNumericMatrix(dims[1], dims[2], mxUINT8_CLASS, mxREAL);
			std::uint8_t* Yj_p = static_cast<std::uint8_t*>(mxGetData(Yj));
			size_t nj = dims[1] * dims[2];

			for (size_t i = 0; i < nj; ++i)
				Yj_p[i] = y_sol[pi + i];
			pi += nj;

			mxSetCell(Ycell, j, Yj);
		}
		assert(pi == y_sol.size());
	}

	// Output recorded iteration data
	if (nlhs >= 2) {
		assert(record_obj.size() == record_time.size());

		plhs[1] = mxCreateNumericMatrix(2, record_obj.size(), mxDOUBLE_CLASS, mxREAL);
		double* R = static_cast<double*>(mxGetData(plhs[1]));
		for (size_t iter = 0; iter < record_obj.size(); ++iter) {
			R[2*iter+0] = record_obj[iter];
			R[2*iter+1] = record_time[iter];
		}
	}

	// Output per-class objective
	if (nlhs >= 3) {
		auto obj = pred.compute_objective_perclass();
		plhs[2] = mxCreateNumericMatrix(1, obj.size(), mxDOUBLE_CLASS, mxREAL);
		double* obj_p = static_cast<double*>(mxGetData(plhs[2]));
		std::copy(obj.cbegin(), obj.cend(), obj_p);
	}

	MatlabCPPExit();
}

