
#ifndef DIRICHLET_H
#define DIRICHLET_H

#include <vector>
#include <iostream>
#include <iomanip>
#include <numeric>
#include <algorithm>
#include <random>
#include <cmath>
#include <cassert>

// Dirichlet distribution
template<typename TFloat>
class dirichlet {
private:
	std::vector<TFloat> alpha;
	TFloat alpha_0;	// sum of alpha elements
	TFloat lB;
	bool symmetric;	// predicate: alpha[ki] == alpha, for all ki

	// Stirling approximation of the log-gamma function, valid only for
	// positive real arguments z > 0.
	//
	// Numerator sequence: http://oeis.org/A046968
	// Denominator sequence: http://oeis.org/A046969
	static TFloat loggamma(TFloat z) {
		// Small arguments are badly approximated so we use the recursive
		// identity loggamma(z) = loggamma(z+1) - log(z).
		if (z < 8.0)
			return loggamma(z+1.0) - std::log(z);

		TFloat z2 = z*z;
		TFloat z4 = z2*z2;
		TFloat z8 = z4*z4;
		TFloat lnGz = (z - 0.5)*std::log(z) - z + 9.189385332046727e-001;

		TFloat it(0);
		it += 1.0/(12.0*z);			// z
		it += -1.0/(360.0*z2*z);	// z3
		it += 1.0/(1260.0*z4*z);	// z5
		it += -1.0/(1680.0*z4*z2*z);// z7
		it += 1.0/(1188.0*z8*z);	// z9
		it += -691.0/(360360.0*z8*z2*z);	// z11

		return (lnGz + it);
	}

	// Return the logarithm of the normalizing constant, the multinomial Beta
	// function
	TFloat log_B(void) const {
		TFloat lb = -loggamma(alpha_0);
		for (size_t ki = 0; ki < alpha.size(); ++ki)
			lb += loggamma(alpha[ki]);

		return lb;
	}

public:
	dirichlet(size_t k) : alpha(k), alpha_0(k),
		lB(std::numeric_limits<TFloat>::signaling_NaN())
	{
		std::fill(alpha.begin(), alpha.end(), 1.0);
		lB = log_B();
	}

	dirichlet(size_t k, TFloat alpha_sym)
		: alpha(k, alpha_sym), alpha_0(static_cast<TFloat>(k)*alpha_sym),
			symmetric(true)
	{
		assert(alpha_sym > 0.0);
		lB = log_B();
	}

	dirichlet(const std::vector<TFloat>& alpha)
		: alpha(alpha),
			alpha_0(std::accumulate(alpha.cbegin(), alpha.cend(), TFloat(0))),
			lB(log_B()), symmetric(false)
	{
		for (size_t ki = 0; ki < alpha.size(); ++ki) {
			assert(alpha[ki] > 0.0);
		}
	}

	// Compute the log-likelihood of a categorial distribution.
	// We must have prob[ki] > 0 for all ki.
	template<typename TProbIterType>
	TFloat log_likelihood(TProbIterType prob) const {
		TFloat ll = 0;
		for (size_t ki = 0; ki < alpha.size(); ++ki) {
			assert(prob[ki] > 0.0);
			ll += (alpha[ki]-1.0)*std::log(prob[ki]);
		}
		return ll - lB;
	}

	// Return the number of bins
	size_t bins(void) const {
		return alpha.size();
	}

	template<typename TVectorType>
	void mean(TVectorType& mean_vec) const {
		mean_vec.resize(alpha.size());
		for (size_t ki = 0; ki < alpha.size(); ++ki)
			mean_vec[ki] = alpha[ki] / alpha_0;
	}

	// Produce a sample from the Dirichlet distribution and return the
	// log-likelihood.
	TFloat simulate(std::vector<TFloat>& x, std::mt19937& eng) const {
		x.resize(alpha.size());
		TFloat xsum(0);
		for (size_t ki = 0; ki < alpha.size(); ++ki) {
			std::gamma_distribution<TFloat> gamma_ki(alpha[ki], TFloat(1));
			x[ki] = gamma_ki(eng);
			xsum += x[ki];
		}
		for (size_t ki = 0; ki < alpha.size(); ++ki)
			x[ki] /= xsum;

		return log_likelihood(x.cbegin());
	}
};

#endif

