
#ifndef LFP_H
#define LFP_H

#include <iostream>
#include <fstream>
#include <iomanip>
#include <algorithm>
#include <numeric>
#include <limits>
#include <vector>
#include <unordered_set>
#include <unordered_map>
#include <cassert>

#include <boost/dynamic_bitset.hpp>

#include "timing.h"

//#define CHECK_PRIMAL
//#define ALPHA

// Linear fractional programming
namespace lfp {

class problem;
std::ostream& operator<<(std::ostream& os, const problem& prob);

/* Solver for
 *
 *     max   (A0+A'*lambda)/(B0+B'*lambda) + (C0+C'*lambda)/(D0+D'*lambda)
 *   lambda
 *    sb.t.  lambda_i >= 0, lambda_i <= 1,  i in {1,2,...,n}.
 *
 * Author: Sebastian Nowozin <Sebastian.Nowozin@microsoft.com>
 * Date: November 2013.
 */
class problem {
public:
	typedef std::vector<double> vector;
	typedef boost::dynamic_bitset<unsigned long> basis_set;

	struct solve_statistics {
	public:
		size_t n;	// Number of decision variables
		double f_xi_opt;	// Optimal objective
		double xi_opt;		// Maximizer
		double fract_opt;	// Fractionality of optimal solution

		double opt_time_ms;	// Solve time to optimal solution (ms)
		double total_time_ms;	// Total solve time (ms)

		// Number of primal pivots to make initial basis optimal
		size_t primal_simplex_init_count;
		// Number of primal simplex calls in main loop
		size_t primal_simplex_count;
		// Number of dual simplex calls in main loop
		size_t dual_simplex_count;

		// Number of times the primal solution was extracted
		size_t primal_extract_count;

		size_t lzez;	// Count of leave-z enter-z
		size_t lzes;	// Count of leave-z enter-s
		size_t lsez;	// Count of leave-s enter-z
		size_t lses;	// Count of leave-s enter-s

		solve_statistics(void)
			: n(0), f_xi_opt(std::numeric_limits<double>::signaling_NaN()),
				xi_opt(std::numeric_limits<double>::signaling_NaN()),
				opt_time_ms(0.0), total_time_ms(0.0),
				primal_simplex_init_count(0), primal_simplex_count(0),
				dual_simplex_count(0), primal_extract_count(0),
				lzez(0), lzes(0), lsez(0), lses(0)
		{
		}
	};

	// Output solver statistics
	friend std::ostream& operator<<(std::ostream& os,
		const solve_statistics& stat)
	{
		std::ios_base::fmtflags os_flags(os.flags());

		os << "OPT f(xi=" << stat.xi_opt << ") = ";
		os << std::fixed << std::scientific << std::setprecision(8);
		os << stat.f_xi_opt;
		os.flags(os_flags);
		os << " after " << stat.opt_time_ms << "ms ("
			<< stat.total_time_ms << "ms total, "
			<< stat.primal_extract_count << " sols); ";
		os << "n " << stat.n << ", ";
		os << std::fixed << std::setprecision(2)
			<< (100.0*stat.fract_opt) << "% fract" << std::endl;
		os.flags(os_flags);
		os << "    SIMPLEX "
			<< stat.primal_simplex_init_count << "i/"
			<< stat.primal_simplex_count << "p/"
			<< stat.dual_simplex_count << "d  PIVOT "
			<< stat.lzez << "zz/" << stat.lzes << "zs/"
			<< stat.lsez << "sz/" << stat.lses << "ss" << std::endl;

		return os;
	}

private:
#define zero_tol 1.0e-13

	// Core problem specification
	size_t n;	// Number of binary variables

	double A0;
	vector A;
	double B0;
	vector B;
	double C0;
	vector C;
	double D0;
	vector D;

	// Charnes-Cooper linearization parameters
	double xi;
	double xi_min;
	double xi_max;

	enum variable_type {
		none = 0,
		z_type = 1,
		s_type = 2,
	};

	class basic_noxi_vector;
	class basic_noxi_densevector;

	// Structure of a solution to Bu = A_{z_j|s_j} is
	//    u.z0: non-zero constant
	//    u.z(i) = 1 + u.z0,   if i==j and s_j is entering
	//    u.z(i) = 0,   if z_i is non-basic
	//    u.z(i) = u.z0 - u.s(k),   if z_i is basic and i == k (z_i should be basic then)
	//    u.z(i) = u.z0,   otherwise
	//    u.s(i) = 1.0 + u.z0,   if i==j and z_j is entering
	//    u.s(i) = u_sk,   if i == k  (z_i, s_i are basic in this case)
	//    u.s(i) = 0,   if s_i is non-basic
	//    u.s(i) = u.z0,   otherwise
	struct Bu_solution {
	private:
		const problem* prob;
		double u_sk;
		size_t k;
		variable_type entering;
		size_t i_enter;

	public:
		friend class problem;
		friend class basic_noxi_vector;
		friend class basic_noxi_densevector;

		double z0;

		double z(size_t i) const {
			if (entering == s_type && i == i_enter)
				return 1.0 + z0;
			if (prob->is_basic_z(i) == false)
				return 0.0;
			if (i == k)
				return z0 - u_sk;

			return z0;
		}
		double s(size_t i) const {
			if (entering == z_type && i == i_enter)
				return 1.0 + z0;
			if (i == k)
				return u_sk;

			if (prob->is_basic_s(i) == false)
				return 0.0;

			return z0;
		}

		void compute_sigma34(double& sigma_3, double& sigma_4) const {
			sigma_3 = z0 * prob->A_basis_sum;
			sigma_4 = z0 * prob->C_basis_sum;

			// Apply corrections
			if (entering == s_type) {
				sigma_3 += prob->A[i_enter];
				sigma_4 += prob->C[i_enter];
			}
			sigma_3 -= u_sk * prob->A[k];
			sigma_4 -= u_sk * prob->C[k];
		}

		Bu_solution(void)
			: prob(0), u_sk(std::numeric_limits<double>::signaling_NaN()),
				k(0), entering(none), i_enter(0)
		{
		}

		Bu_solution(const problem* prob, double u_z0, double u_sk, size_t k,
			variable_type entering = none, size_t i_enter = 0)
			: prob(prob), u_sk(u_sk), k(k), entering(entering), i_enter(i_enter),
				z0(u_z0)
		{
		}
	};

	// Vector class to represent z_noxi and s_noxi in semi-sparse form and
	// allow for efficient O(1) updates
	class basic_noxi_vector {
	private:
		variable_type vt;	// represent s_i or z_i
		double offset;
		std::unordered_map<size_t, double> zs_noxi_raw;

		void check_zero(size_t i) {
			auto bi = zs_noxi_raw.find(i);
			if (bi == zs_noxi_raw.end())
				return;

			if (std::fabs(bi->second) <= 1.0e-13)
				zs_noxi_raw.erase(bi);
		}

	public:
		basic_noxi_vector(void)
			: vt(none), offset(std::numeric_limits<double>::signaling_NaN())
		{
		}

		basic_noxi_vector(variable_type vt, double offset)
			: vt(vt), offset(offset)
		{
		}

		// i must be basic!
		double operator()(size_t i) const {
			auto yi = zs_noxi_raw.find(i);
			if (yi == zs_noxi_raw.end())
				return offset;

			return yi->second + offset;
		}

		// Efficiently compute
		//    sigma_1 = \sum_{z_i basic} A[i] z_noxi[i]
		//    sigma_2 = \sum_{z_i basic} C[i] z_noxi[i]
		void compute_sigma12(const problem* prob,
			double& sigma_1, double& sigma_2) const
		{
			// Dense part
			sigma_1 = prob->A_basis_sum*offset;
			sigma_2 = prob->C_basis_sum*offset;

			for (auto zsi = zs_noxi_raw.cbegin(); zsi != zs_noxi_raw.cend(); ++zsi) {
				if (prob->is_basic_z(zsi->first)) {
					sigma_1 += zsi->second * prob->A[zsi->first];
					sigma_2 += zsi->second * prob->C[zsi->first];
				}
			}
		}

		// += scale * v
		void add(double scale, const Bu_solution& v) {
			// u.z(i) = u.z0,   otherwise
			// u.s(i) = u.z0,   otherwise
			offset += scale*v.z0;

			if (vt == z_type) {
				// u.z(i) = 1 + u.z0,   if i==j and s_j is entering
				if (v.entering == s_type) {
					zs_noxi_raw[v.i_enter] += scale;
					check_zero(v.i_enter);
				}

				// u.z(i) = u.z0 - u.s(k),   if z_i is basic and i == k
				zs_noxi_raw[v.k] -= scale * v.u_sk;
				check_zero(v.k);
			} else if (vt == s_type) {
				// u.s(i) = 1.0 + u.z0,   if i==j and z_j is entering
				if (v.entering == z_type) {
					zs_noxi_raw[v.i_enter] += scale;
					check_zero(v.i_enter);
				}

				// u.s(i) = u_sk,   if i == k  (z_i, s_i are basic in this case)
				zs_noxi_raw[v.k] += scale*(v.u_sk - v.z0);
				check_zero(v.k);
			} else {
				assert(0);
			}
		}

		// i must be basic
		void fix_value(size_t i, double value) {
			double current = operator()(i);
			if (std::fabs(current - value) <= 1.0e-14)
				return;

			zs_noxi_raw[i] += value - current;
			check_zero(i);
		}

		// Force an element out of the basis, conserving semi-sparsity
		void make_nonbasic(size_t i) {
			zs_noxi_raw.erase(i);
		}
	};

	// Vector class to represent z_noxi and s_noxi in semi-sparse form and
	// allow for efficient dense O(1) updates
	class basic_noxi_densevector {
	private:
		variable_type vt;	// represent s_i or z_i
		double offset;
		std::vector<double> zs_noxi_raw;

	public:
		basic_noxi_densevector()
			: vt(none), offset(std::numeric_limits<double>::signaling_NaN())
		{
		}

		basic_noxi_densevector(size_t n, variable_type vt, double offset)
			: vt(vt), offset(offset), zs_noxi_raw(n, 0.0)
		{
		}

		// i must be basic!
		double operator()(size_t i) const {
			return zs_noxi_raw[i] + offset;
		}

		// += scale * v
		void add(double scale, const Bu_solution& v) {
			// u.z(i) = u.z0,   otherwise
			// u.s(i) = u.z0,   otherwise
			offset += scale*v.z0;

			if (vt == z_type) {
				// u.z(i) = 1 + u.z0,   if i==j and s_j is entering
				if (v.entering == s_type)
					zs_noxi_raw[v.i_enter] += scale;

				// u.z(i) = u.z0 - u.s(k),   if z_i is basic and i == k
				zs_noxi_raw[v.k] -= scale * v.u_sk;
			} else if (vt == s_type) {
				// u.s(i) = 1.0 + u.z0,   if i==j and z_j is entering
				if (v.entering == z_type)
					zs_noxi_raw[v.i_enter] += scale;

				// u.s(i) = u_sk,   if i == k  (z_i, s_i are basic in this case)
				zs_noxi_raw[v.k] += scale*(v.u_sk - v.z0);
			} else {
				assert(0);
			}
		}

		// i must be basic
		void fix_value(size_t i, double value) {
			double current = operator()(i);
			if (std::fabs(current - value) <= 1.0e-14)
				return;

			zs_noxi_raw[i] += value - current;
		}

		// Force an element out of the basis, conserving semi-sparsity
		void make_nonbasic(size_t i) {
			zs_noxi_raw[i] = -offset;	// set to zero
		}
	};

#ifdef REDCOST_TODO
	class redcost_vector {
	private:
		const problem* prob;
		variable_type vtype;

		double tau1;
		double tau2;

		// Maintain per-element shifts
		typedef boost::dynamic_bitset<unsigned long> dset;
		dset has_shift;
		std::vector<double> shift;

	public:
		redcost_vector(const problem* prob, size_t n, variable_type vtype)
			: prob(prob), vtype(vtype), tau1(0.0), tau2(0.0),
				has_shift(n), shift(n, 0.0)
		{
		}

		double operator()(size_t i) const {
#if 1
			if (vtype == z_type && prob->is_basic_z(i))
				return 0.0;
			if (vtype == s_type && prob->is_basic_s(i))
				return 0.0;
#endif

			double v = tau1*prob->D[i] + tau2*prob->B[i];
			return has_shift.test(i) ? (v + shift[i]) : v;
		}

		double get_raw_value(size_t i) const {
			double v = tau1*prob->D[i] + tau2*prob->B[i];
			return has_shift.test(i) ? (v + shift[i]) : v;
		}

		void add_tau(double tau1_add, double tau2_add) {
			tau1 += tau1_add;
			tau2 += tau2_add;
		}

		void add(size_t i, double value) {
			if (value == 0.0)
				return;

//			if (has_shift.test(i) == false)
//				shift[i] = 0.0;
			shift[i] += value;
			has_shift.set(i);
		}

		void make_basic(size_t i) {
//			add(i, -tau1*prob->D[i] -tau2*prob->B[i]);
		}
		void make_nonbasic(size_t i) {
//			add(i, -tau1*prob->D[i] -tau2*prob->B[i]);
//			add(i, -get_raw_value(i));
		}
	};
#endif

	// New basis representation
	basis_set basis_z;
	bool is_basic_z(size_t i) const { return basis_z.test(i); };
	basis_set basis_s;
	bool is_basic_s(size_t i) const { return basis_s.test(i); };
	size_t k_shared;	// z_k and s_k are in basis

	// Primal solution representation
#ifdef CHECK_PRIMAL
	double z0_xi;
	double z0_noxi;
#endif
	double z0_xi2;
	double z0_noxi2;
	double z0(void) const { return z0_noxi2 + xi*z0_xi2; }
	double z0(double at_xi) const { return z0_noxi2 + at_xi*z0_xi2; }

#ifdef CHECK_PRIMAL
	vector z_xi;
	vector z_noxi;
#endif
	Bu_solution z_xi2;
	basic_noxi_vector z_noxi2;
	double z(size_t i) const { return z_noxi2(i) + xi*z_xi2.z(i); }
	double z(size_t i, double at_xi) const { return z_noxi2(i) + at_xi*z_xi2.z(i); }
#ifdef CHECK_PRIMAL
	vector s_xi;
	vector s_noxi;
#endif
	Bu_solution s_xi2;
	basic_noxi_densevector s_noxi2;
	double s(size_t i) const { return s_noxi2(i) + xi*s_xi2.s(i); }

	double cost_z0_xi(void) const { return -C0; };
	double cost_z0_noxi(void) const { return -A0; };
	double cost_z0(void) const { return cost_z0_xi()*xi + cost_z0_noxi(); };
	double cost_z_xi(size_t i) const { return -C[i]; };
	double cost_z_noxi(size_t i) const { return -A[i]; };
	double cost_z(size_t i) const { return cost_z_xi(i)*xi + cost_z_noxi(i); };
	double cost_s_xi(size_t i) const { return 0.0; };
	double cost_s_noxi(size_t i) const { return 0.0; };
	double cost_s(size_t i) const { return 0.0; };

	// Reduced cost representation
	double redcost_z0_xi;
	double redcost_z0_noxi;
	double redcost_z0(void) const {
		return cost_z0() - xi*redcost_z0_xi - redcost_z0_noxi; };
	vector redcost_z_xi;
	vector redcost_z_noxi;
#ifdef REDCOST_TODO
	redcost_vector redcost_z_xi2;
	redcost_vector redcost_z_noxi2;
#endif
	double redcost_z(size_t i) const {
		return cost_z(i) - xi*redcost_z_xi[i] - redcost_z_noxi[i]; };
	vector redcost_s_xi;
	vector redcost_s_noxi;
#ifdef REDCOST_TODO
	redcost_vector redcost_s_xi2;
	redcost_vector redcost_s_noxi2;
#endif
	double redcost_s(size_t i) const {
		return cost_s(i) - xi*redcost_s_xi[i] - redcost_s_noxi[i]; };

	// Basis sums that are needed often
	double D0_basis_sum;
	double B0_basis_sum;
	double A_basis_sum;
	double C_basis_sum;

	// Compute primal infeasibility measure
	double pinf(void) const {
		double inf = std::max(0.0, -z0());

		for (size_t bzi = basis_z.find_first();
			bzi != basis_set::npos; bzi = basis_z.find_next(bzi))
		{
			inf += std::max(0.0, -z(bzi));
		}
		for (size_t bsi = basis_s.find_first();
			bsi != basis_set::npos; bsi = basis_s.find_next(bsi))
		{
			inf += std::max(0.0, -s(bsi));
		}

		return inf;
	}

	// Solve [Q11 Q12; Q21 Q22] \ [s;t] assuming full rank
	static std::pair<double, double> solve_2by2(double Q11, double Q12,
		double Q21, double Q22, double s, double t) {
		double ac = Q11 / Q21;
		double v = (t*ac - s) / (Q22*ac - Q12);
		double u = (s - Q12*v) / Q11;

		return std::pair<double, double>(u, v);
	}

	// SLOW version of primal objective
	double pobj_lp_slow(void) const {
		double obj = cost_z0() * z0();
		for (size_t bzi = basis_z.find_first();
			bzi != basis_set::npos; bzi = basis_z.find_next(bzi))
		{
			obj += cost_z(bzi) * z(bzi);
		}

		return obj;
	}

	// Compute the polynomial/rational expansion of the LP objective function
	// in xi as
	//
	//   f(xi) = (t0 + t1*xi + t2*xi^2)/xi,
	//
	// This expansion is valid in the current xi_range only.
	// Return tuple (t0,t1,t2).
	std::tuple<double, double, double> compute_quadratic_objective(void) const
	{
		double sigma_1;
		double sigma_2;
		z_noxi2.compute_sigma12(this, sigma_1, sigma_2);

		double sigma_3;
		double sigma_4;
		z_xi2.compute_sigma34(sigma_3, sigma_4);

		double t0 = sigma_1 + A0*z0_noxi2;
		double t1 = C0*z0_noxi2 + A0*z0_xi2 + sigma_2 + sigma_3;
		double t2 = C0*z0_xi2 + sigma_4;

		return std::tuple<double, double, double>(t0, t1, t2);
	}

	double compute_objective_maximizer(
		const std::tuple<double, double, double>& t,
		double xi_lb, double xi_ub, double& f_max) const
	{
		double t0 = std::get<0>(t);
		double t1 = std::get<1>(t);
		double t2 = std::get<2>(t);

		// Compute function values at boundaries
		double f_lb = t0/xi_lb + t1 + t2*xi_lb;
		double f_ub = t0/xi_ub + t1 + t2*xi_ub;

		// Compute critical point
		double xi_c = std::sqrt(t0 / t2);
		// Check it is in valid range
		if (xi_c >= xi_lb && xi_c <= xi_ub) {
			// Check it is a maximum
			double xi_c_d2 = 2.0 * t0 / (xi_c*xi_c*xi_c);
			if (xi_c_d2 < 0.0) {
				double f_xi_c = t0/xi_c + t1 + t2*xi_c;
				assert(f_xi_c >= f_lb && f_xi_c >= f_ub);

				f_max = f_xi_c;
				return xi_c;
			}
		}

		if (f_lb >= f_ub) {
			f_max = f_lb;
			return xi_lb;
		}

		f_max = f_ub;
		return xi_ub;
	}

	double pobj_lp(void) const {
		double sigma_1;
		double sigma_2;
		z_noxi2.compute_sigma12(this, sigma_1, sigma_2);

		double sigma_3;
		double sigma_4;
		z_xi2.compute_sigma34(sigma_3, sigma_4);

#ifdef VERIFY_COST
		// Optional verification
		double sigma_1c = 0.0;
		double sigma_2c = 0.0;
		double sigma_3c = 0.0;
		double sigma_4c = 0.0;

		for (size_t bzi = basis_z.find_first();
			bzi != basis_set::npos; bzi = basis_z.find_next(bzi))
		{
			sigma_1c += A[bzi] * z_noxi2(bzi);
			sigma_2c += C[bzi] * z_noxi2(bzi);
			sigma_3c += A[bzi] * z_xi2.z(bzi);
			sigma_4c += C[bzi] * z_xi2.z(bzi);
		}
		assert(std::fabs(sigma_1 - sigma_1c) <= 1.0e-13);
		assert(std::fabs(sigma_2 - sigma_2c) <= 1.0e-13);
		assert(std::fabs(sigma_3 - sigma_3c) <= 1.0e-13);
		assert(std::fabs(sigma_4 - sigma_4c) <= 1.0e-13);
#endif

		double obj_lp = -A0*z0_noxi2 - A0*xi*z0_xi2 - xi*C0*z0_noxi2 - xi*xi*C0*z0_xi2;
		obj_lp += -sigma_1;
		obj_lp += -xi*(sigma_2 + sigma_3);
		obj_lp += -xi*xi*sigma_4;

		return obj_lp;
	}

	double pobj(void) const {
		return -pobj_lp() / xi;
	}

	enum limiting_type {
		dual_feasibility = 1,
		primal_feasibility = 2,
	};

	// Compact representation of a solution to pB=e
	struct pB_solution {
	private:
		const problem& prob;
		variable_type vtype;
		size_t j;

	public:
		friend class problem;

		double alpha;
		double beta;

		// gamma_i = 1,  for s_type and i == j, z_i does not have to be basic
		// gamma_i = 0,  for non-basic z_i
		// gamma_i = -alpha*D[i]-beta*B[i],  for basic z_i, i != j,
		// gamma_i = -alpha*D[i]-beta*B[i] + 1,  for basic z_i, i == j
		double gamma(size_t i) const {
			if (vtype == s_type && j == i)
				return 1.0;

			double res = prob.is_basic_z(i) ?
				(-prob.D[i]*alpha - prob.B[i]*beta) : 0.0;
			if (vtype == z_type && j == i) {
				res += 1.0;
			}

			return res;
		}

#if 0
		// sum(gamma)
		double gamma_sum(void) const {
			double sum = 0.0;
			for (auto bzi = prob.basis_z.cbegin(); bzi != prob.basis_z.cend(); ++bzi)
				sum += gamma(*bzi);
			if (prob.is_basic_z(j) == false)
				sum += 1.0;

			return sum;
		}
#endif

		pB_solution(const problem& prob, variable_type vtype, size_t j,
			double alpha, double beta)
			: prob(prob), vtype(vtype), j(j), alpha(alpha), beta(beta)
		{
		}
	};
	// Solve pB = e_li, ported from qsolve_pB_e.m
	pB_solution solve_pB_e(variable_type vtype, size_t j) const {
		double Q11 = D0_basis_sum;
		double Q12 = B0_basis_sum;

		double b2 = 0.0;
		if (vtype == z_type) {
			assert(j < n && is_basic_z(j));

			if (k_shared == j)
				b2 = 1.0;
		} else {
			assert(vtype == s_type);
			assert(j < n && is_basic_s(j));

			if (k_shared == j) {
				b2 = -1.0;
				Q11 -= D[k_shared];
				Q12 -= B[k_shared];
			}
		}
		auto ab = solve_2by2(Q11, Q12, D[k_shared], B[k_shared], 1, b2);

		return pB_solution(*this, vtype, j, ab.first, ab.second);
	}

	// Solve Bu = A_{z_j}, ported from qsolve_Bu_Azj2.m
	Bu_solution solve_Bu_Azj(size_t z_enter) const {
		double Q11 = D0_basis_sum;
		double Q21 = B0_basis_sum;
		auto ut = solve_2by2(Q11, -D[k_shared], Q21, -B[k_shared],
			D[z_enter], B[z_enter]);

		return Bu_solution(this, ut.first, ut.second, k_shared, z_type, z_enter);
	}

	// Solve Bu = A_{s_j}, ported from qsolve_Bu_Asj2.m
	Bu_solution solve_Bu_Asj(size_t s_enter) const {
		double Q11 = D0_basis_sum;
		double Q21 = B0_basis_sum;
		double rhs_0 = 0.0;
		double rhs_1 = 0.0;

		if (is_basic_z(s_enter)) {
			rhs_0 = -D[s_enter];
			rhs_1 = -B[s_enter];
		}
		auto ut = solve_2by2(Q11, -D[k_shared], Q21, -B[k_shared], rhs_0, rhs_1);

		return Bu_solution(this, ut.first, ut.second, k_shared, s_type, s_enter);
	}

	// Solve Bu = [1; zeros(m-1,1)] and get xi-dependent vector.
	// Ported from qsolve_Bu_1.m
	Bu_solution solve_Bu_1(void) const {
		double Q11 = D0_basis_sum;
		double Q21 = B0_basis_sum;
		auto ut = solve_2by2(Q11, -D[k_shared], Q21, -B[k_shared], 0.0, 1.0);

		return Bu_solution(this, ut.first, ut.second, k_shared);
	}

	// XXX: This is the only place besides the initialization where the basis
	// is changed.
	void update_basis_heading(variable_type leaving, size_t i_leave,
		variable_type entering, size_t i_enter)
	{
		// Update basis indices
		if (leaving == z_type) {
			// z_{i_leave} leaves the basis
			basis_z[i_leave] = false;
#ifdef CHECK_PRIMAL
			z_xi[i_leave] = 0.0;
			z_noxi[i_leave] = 0.0;
#endif
			z_noxi2.make_nonbasic(i_leave);

#ifdef REDCOST_TODO
			redcost_z_xi2.make_nonbasic(i_leave);
			redcost_z_noxi2.make_nonbasic(i_leave);
#endif

			D0_basis_sum -= D[i_leave];
			B0_basis_sum -= B[i_leave];
			A_basis_sum -= A[i_leave];
			C_basis_sum -= C[i_leave];
		} else if (leaving == s_type) {
			// s_{i_leave} leaves the basis
			basis_s[i_leave] = false;
#ifdef CHECK_PRIMAL
			s_xi[i_leave] = 0.0;
			s_noxi[i_leave] = 0.0;
#endif
			s_noxi2.make_nonbasic(i_leave);
#ifdef REDCOST_TODO
			redcost_s_xi2.make_nonbasic(i_leave);
			redcost_s_noxi2.make_nonbasic(i_leave);
#endif
		} else {
			assert(0);
		}
		if (entering == z_type) {
			// z_{i_enter} enters the basis
			assert(is_basic_z(i_enter) == false);
			basis_z[i_enter] = true;
			if (is_basic_s(i_enter))
				k_shared = i_enter;	// z_k, s_k both in basis

			D0_basis_sum += D[i_enter];
			B0_basis_sum += B[i_enter];
			A_basis_sum += A[i_enter];
			C_basis_sum += C[i_enter];

#ifdef REDCOST_TODO
			redcost_z_xi2.make_basic(i_enter);
			redcost_z_noxi2.make_basic(i_enter);
#endif
		} else if (entering == s_type) {
			// s_{i_enter} enters the basis
			assert(is_basic_s(i_enter) == false);
			basis_s[i_enter] = true;
			if (is_basic_z(i_enter))
				k_shared = i_enter;	// z_k, s_k both in basis

#ifdef REDCOST_TODO
			redcost_s_xi2.make_basic(i_enter);
			redcost_s_noxi2.make_basic(i_enter);
#endif
		} else {
			assert(0);
		}
	}

#ifdef CHECK_PRIMAL
	void check_primal(void) const {
		assert(std::fabs(z0_noxi - z0_noxi2) <= 1.0e-14);
		assert(std::fabs(z0_xi - z0_xi2) <= 1.0e-14);

		for (size_t bzi = basis_z.find_first();
			bzi != basis_set::npos; bzi = basis_z.find_next(bzi))
		{
			if (std::fabs(z_noxi[bzi] - z_noxi2(bzi)) > 1.0e-14) {
				std::cout << "z_" << bzi << " disagree:" << std::endl;
				std::cout << "   z_noxi[" << bzi << "] " << z_noxi[bzi] << std::endl;
				std::cout << "   z_noxi2[" << bzi << "] " << z_noxi2(bzi) << std::endl;
				std::cout << "   xi " << xi << std::endl;
			}
			assert(std::fabs(z_noxi[bzi] - z_noxi2(bzi)) <= 1.0e-14);
		}
		for (size_t i = 0; i < n; ++i) {
			assert(std::fabs(z_xi2.z(i) - z_xi[i]) <= 1.0e-14);
		}
		for (size_t bsi = basis_s.find_first();
			bsi != basis_set::npos; bsi = basis_s.find_next(bsi))
		{
			assert(std::fabs(s_noxi[bsi] - s_noxi2(bsi)) <= 1.0e-14);
		}
	}
#endif

	// Update primal representation of xi-decoupled solution
	// This is the only place where z0, z_i, s_i are modified.
	void update_basis_solution(const Bu_solution& u, const Bu_solution& v,
		double theta, variable_type entering, size_t i_enter)
	{
#ifdef CHECK_PRIMAL
		// Update z0
		double z0_prev = z0();
		double z0_new = z0_prev - theta*u.z0;
		z0_xi = v.z0;
		z0_noxi = z0_new - xi*v.z0;
#endif

		// O(1) update for z0_noxi, z0_xi
		z0_noxi2 += xi*z0_xi2;
		z0_noxi2 -= theta*u.z0;
		z0_noxi2 -= xi*v.z0;
		z0_xi2 = v.z0;

#ifdef CHECK_PRIMAL
		// Update z
		for (size_t bzi = basis_z.find_first();
			bzi != basis_set::npos; bzi = basis_z.find_next(bzi))
		{
			double zi_prev = z(bzi);
			double zi_new = zi_prev - theta*u.z(bzi);
			if (entering == z_type && bzi == i_enter)
				zi_new = theta;

			z_xi[bzi] = v.z(bzi);
			z_noxi[bzi] = zi_new - xi*v.z(bzi);
		}
#endif

		// O(1) update for z_noxi, z_xi
		z_noxi2.add(xi, z_xi2);
		z_noxi2.add(-theta, u);
		z_noxi2.add(-xi, v);

		if (entering == z_type) {
			assert(is_basic_z(i_enter));	// already in basis heading
			z_noxi2.fix_value(i_enter, theta - xi*v.z(i_enter));
		}
		z_xi2 = v;

#ifdef CHECK_PRIMAL
		// Update s
		for (size_t bsi = basis_s.find_first(); bsi != basis_set::npos;
			bsi = basis_s.find_next(bsi))
		{
			double si_prev = s(bsi);
			double si_new = si_prev - theta*u.s(bsi);
			if (entering == s_type && bsi == i_enter)
				si_new = theta;

			s_xi[bsi] = v.s(bsi);
			s_noxi[bsi] = si_new - xi*v.s(bsi);
		}
#endif

		// O(1) update for s_noxi, s_xi
		s_noxi2.add(xi, s_xi2);
		s_noxi2.add(-theta, u);
		s_noxi2.add(-xi, v);
		if (entering == s_type)
			s_noxi2.fix_value(i_enter, theta - xi*v.s(i_enter));
		s_xi2 = v;

#ifdef CHECK_PRIMAL
		check_primal();
#endif
	}

	void update_reduced_costs(const pB_solution& p, variable_type entering,
		size_t i_enter)
	{
		// Decoupled reduced costs by xi-dependence
		double rc_xi = 0;
		double rc_noxi = 0;
		if (entering == z_type) {
			rc_xi = cost_z_xi(i_enter) - redcost_z_xi[i_enter];
			rc_noxi = cost_z_noxi(i_enter) - redcost_z_noxi[i_enter];
		} else if (entering == s_type) {
			rc_xi = cost_s_xi(i_enter) - redcost_s_xi[i_enter];
			rc_noxi = cost_s_noxi(i_enter) - redcost_s_noxi[i_enter];
		} else {
			assert(0);
		}

#if 0
		// (We do not need to track redcost_z0)
		// Update reduced costs for z0
		double pAz0 = p.alpha*D0 + p.beta*B0 - p.gamma_sum();
		redcost_z0_xi += rc_xi * pAz0;
		redcost_z0_noxi += rc_noxi * pAz0;
#endif

#if 0
		// Update reduced costs for z_i, s_i
#if 0
		std::cout << "p.alpha " << p.alpha << "  p.beta " << p.beta << std::endl;
		std::cout << "rc_xi " << rc_xi << "  rc_noxi " << rc_noxi << std::endl;
#endif

		// Working version (simple), but faster version below
		for (size_t i = 0; i < n; ++i) {
			// Basic reduced costs stay at zero
			double pAzi = p.alpha*D[i] + p.beta*B[i] + p.gamma(i);

#if 1
//			if (i == 126) {
			std::cout << "z_" << i << " is " << (is_basic_z(i) ? "BASIC" : "NONBASIC") << std::endl;
			std::cout << " PRE redcost_z_xi2(" << i << ") " << redcost_z_xi2(i)
				<< "  redcost_z_noxi2(" << i << ") " << redcost_z_noxi2(i)
				<< std::endl;
			std::cout << " PRE redcost_z_xi2-raw(" << i << ") " << redcost_z_xi2.get_raw_value(i)
				<< "  redcost_z_noxi2-raw(" << i << ") " << redcost_z_noxi2.get_raw_value(i)
				<< std::endl;
			std::cout << " PRE redcost_z_xi[" << i << "] " << redcost_z_xi[i] << std::endl;
			std::cout << "     redcost_z_noxi[" << i << "] " << redcost_z_noxi[i] << std::endl;
			std::cout << "     A rc_xi * pAzi = " << (rc_xi * pAzi) << std::endl;
//			}
#endif

			// Update reduced costs of z_i
			redcost_z_xi[i] += rc_xi * pAzi;
			redcost_z_noxi[i] += rc_noxi * pAzi;

#if 1
//			if (i == 126) {
			std::cout << "POST redcost_z_xi2(" << i << ") " << redcost_z_xi2(i)
				<< "  redcost_z_noxi2(" << i << ") " << redcost_z_noxi2(i)
				<< std::endl;
			std::cout << "POST redcost_z_xi2-raw(" << i << ") " << redcost_z_xi2.get_raw_value(i)
				<< "  redcost_z_noxi2-raw(" << i << ") " << redcost_z_noxi2.get_raw_value(i)
				<< std::endl;
			std::cout << "POST redcost_z_xi[" << i << "] " << redcost_z_xi[i] << std::endl;
			std::cout << "     redcost_z_noxi[" << i << "] " << redcost_z_noxi[i] << std::endl;
//			}
#endif

			double pAsi = p.gamma(i);

#if 1
//			if (i == 299) {
			std::cout << "s_" << i << " is " << (is_basic_s(i) ? "BASIC" : "NONBASIC") << std::endl;
			std::cout << " PRE redcost_s_xi2(" << i << ") " << redcost_s_xi2(i)
				<< "  redcost_s_noxi2(" << i << ") " << redcost_s_noxi2(i)
				<< std::endl;
			std::cout << " PRE redcost_s_xi2-raw(" << i << ") " << redcost_s_xi2.get_raw_value(i)
				<< "  redcost_s_noxi2-raw(" << i << ") " << redcost_s_noxi2.get_raw_value(i)
				<< std::endl;
			std::cout << " PRE redcost_s_xi[" << i << "] " << redcost_s_xi[i] << std::endl;
			std::cout << "     redcost_s_noxi[" << i << "] " << redcost_s_noxi[i] << std::endl;
			std::cout << "     A rc_xi * pAsi = " << (rc_xi * pAsi) << std::endl;
//			}
#endif

			// Update reduced costs of s_i
			redcost_s_xi[i] += rc_xi * pAsi;
			redcost_s_noxi[i] += rc_noxi * pAsi;

#if 1
//			if (i == 299) {
			std::cout << "POST redcost_s_xi[" << i << "] " << redcost_s_xi[i] << std::endl;
			std::cout << "     redcost_s_noxi[" << i << "] " << redcost_s_noxi[i] << std::endl;
//			}
#endif
		}
#endif

		// O(1) reduced cost updates
#if 0
		redcost_z_xi2.add_tau(rc_xi*p.alpha, rc_xi*p.beta);
		redcost_z_noxi2.add_tau(rc_noxi*p.alpha, rc_noxi*p.beta);
		redcost_z_xi2.add(p.j, rc_xi*p.gamma(p.j));
		redcost_z_noxi2.add(p.j, rc_noxi*p.gamma(p.j));

		redcost_s_xi2.add_tau(-rc_xi*p.alpha, -rc_xi*p.beta);
		redcost_s_noxi2.add_tau(-rc_noxi*p.alpha, -rc_noxi*p.beta);
		// NOT REQUIRED? TODO
		// (or only if it is non-basic?)
		redcost_s_xi2.add(p.j, rc_xi*p.gamma(p.j));
		redcost_s_noxi2.add(p.j, rc_noxi*p.gamma(p.j));
#endif

#if 0
		if (entering == z_type) {
			// For all basic z_i, i != j, we have
			//    gamma(i) = -alpha D[i] -beta B[i],
			// hence the update to the reduced cost for z_i is zero.  For the
			// non-basic z_i gamma is zero, hence p A(:,z_i) is equal to alpha
			// D[i] + beta B[i], which we add.
			redcost_z_xi2.add_tau(rc_xi*p.alpha, rc_xi*p.beta);
			redcost_z_noxi2.add_tau(rc_noxi*p.alpha, rc_noxi*p.beta);

			// For basic z_j, we have
			//    gamma(j) = -alpha D[j] -beta B[j] + 1,
			// hence the update to the reduced costs of z_j is 1.
			// TODO: really add this to the shift
			redcost_z_xi2.add(p.j, rc_xi*1.0);
			redcost_z_noxi2.add(p.j, rc_noxi*1.0);

			// For all basic z_i, i != j, we have
			//    gamma(i) = -alpha D[i] -beta B[i],
			// hence the update to the reduced costs of s_i is just gamma.
			// For the non-basic z_i, i != j, we have
			//    gamma(i) = 0.
			redcost_s_xi2.add_tau(-rc_xi*p.alpha, -rc_xi*p.beta);
			redcost_s_noxi2.add_tau(-rc_noxi*p.alpha, -rc_noxi*p.beta);

			// For basic z_j, we have
			//    gamma(j) = -alpha D[j] -beta B[j] + 1,
			// hence the update to the reduced costs of s_j is 1
			redcost_s_xi2.add(p.j, rc_xi*1.0);
			redcost_s_noxi2.add(p.j, rc_noxi*1.0);
		} else if (entering == s_type) {
			// For all basic z_i, i != j, we have
			//    gamma(i) = -alpha D[i] -beta B[i],
			// hence the update to the reduced cost for z_i is zero.  For the
			// non-basic z_i gamma is zero, hence p A(:,z_i) is again equal to
			// alpha D[i] + beta B[i], which we add.
			redcost_z_xi2.add_tau(rc_xi*p.alpha, rc_xi*p.beta);
			redcost_z_noxi2.add_tau(rc_noxi*p.alpha, rc_noxi*p.beta);

			// For basic z_j, we have
			//    gamma(j) = 1,
			// hence the update to the reduced costs of z_j is 1, but we
			// already added the wrong value above, hence subtract it.
			double rc_zj_add = -p.alpha*D[p.j]-p.beta*B[p.j]+1.0;
			redcost_z_xi2.add(p.j, rc_xi*rc_zj_add);
			redcost_z_noxi2.add(p.j, rc_noxi*rc_zj_add);

			// For all basic z_i, i != j, we have
			//    gamma(i) = -alpha D[i] -beta B[i],
			// hence the update to the reduced cost for s_i is just gamma.
			// For the non-basic z_i, i != j, we have
			//    gamma(i) = 0.
			redcost_s_xi2.add_tau(-rc_xi*p.alpha, -rc_xi*p.beta);
			redcost_s_noxi2.add_tau(-rc_noxi*p.alpha, -rc_noxi*p.beta);

			// For i == j, we have
			//    gamma(j) = 1.
			// hence the update to the reduced costs of s_j is 1, but we
			// already added the wrong value, hence subtract it.
			redcost_s_xi2.add(p.j, rc_xi*rc_zj_add);
			redcost_s_noxi2.add(p.j, rc_noxi*rc_zj_add);
		}
#endif

#if 0
		size_t i = 299;
		std::cout << "POST redcost_s_xi2(" << i << ") " << redcost_s_xi2(i)
			<< "  redcost_s_noxi2(" << i << ") " << redcost_s_noxi2(i)
			<< std::endl;
		std::cout << "POST redcost_s_xi2-raw(" << i << ") " << redcost_s_xi2.get_raw_value(i)
			<< "  redcost_s_noxi2-raw(" << i << ") " << redcost_s_noxi2.get_raw_value(i)
			<< std::endl;
#endif

#if 1
		// Slightly optimized loop (verified, works)
		for (size_t i = 0; i < n; ++i) {
			bool basic_z = is_basic_z(i);
			if (basic_z || i == p.j) {
				double pAsi = p.gamma(i);
				redcost_s_xi[i] += rc_xi * pAsi;
				redcost_s_noxi[i] += rc_noxi * pAsi;
			}
			if (basic_z == false || i == p.j) {
				double pAzi = p.alpha*D[i] + p.beta*B[i] + p.gamma(i);
				redcost_z_xi[i] += rc_xi * pAzi;
				redcost_z_noxi[i] += rc_noxi * pAzi;
			}
		}
#endif

#if 0
		// CHECK
#if 1
		std::cout << "p.j " << p.j << std::endl;
		std::cout << "p.vtype " << p.vtype << std::endl;
		std::cout << "xi " << xi << std::endl;
#endif
		for (size_t i = 0; i < n; ++i) {
//			if (is_basic_z(i) == false) {
#if 1
//				if (i == 126) {
				std::cout << (is_basic_z(i) ? "BASIC" : "NONBASIC") << std::endl;
				std::cout << "  cost_z(i) " << cost_z(i) << std::endl;
				std::cout << "rc_z_xi2(" << i << ") " << redcost_z_xi2(i)
					<< "  rc_z_xi[" << i << "] " << redcost_z_xi[i] << std::endl;
//				}
#endif
				assert(std::fabs(redcost_z_xi2(i) - redcost_z_xi[i]) <= 1.0e-10);
#if 1
				std::cout << "rc_z_noxi2(" << i << ") " << redcost_z_noxi2(i)
					<< "  rc_z_noxi[" << i << "] " << redcost_z_noxi[i] << std::endl;
#endif
				assert(std::fabs(redcost_z_noxi2(i) - redcost_z_noxi[i]) <= 1.0e-10);
//			}

//			if (is_basic_s(i) == false) {
#if 1
//				if (i == 299) {
				std::cout << (is_basic_s(i) ? "BASIC" : "NONBASIC") << std::endl;
				std::cout << "rc_s_xi2(" << i << ") " << redcost_s_xi2(i)
					<< "  rc_s_xi[" << i << "] " << redcost_s_xi[i] << std::endl;
//				}
#endif
				assert(std::fabs(redcost_s_xi2(i) - redcost_s_xi[i]) <= 1.0e-10);
#if 1
				std::cout << "rc_s_noxi2(" << i << ") " << redcost_s_noxi2(i)
					<< "  rc_s_noxi[" << i << "] " << redcost_s_noxi[i] << std::endl;
#endif
				assert(std::fabs(redcost_s_noxi2(i) - redcost_s_noxi[i]) <= 1.0e-10);
//			}
		}
#endif
	}

	// Ported from qupdate_primal_solution.m
	void update_primal_solution(const Bu_solution& u, double theta,
		variable_type leaving, size_t i_leave,
		variable_type entering, size_t i_enter)
	{
		// Update basis heading
		update_basis_heading(leaving, i_leave, entering, i_enter);

		// After the basis heading update the solution is no longer valid
		// until after we have called update_basis_solution(!)

		// Update basic solution, decomposed by xi-dependence
		// This needs to happen after the basis update because we need the new
		// basis for computing v.
		auto v = solve_Bu_1();
		update_basis_solution(u, v, theta, entering, i_enter);

		// Dual simplex reduced cost updates due to [Tomlin1974]
		auto p = solve_pB_e(entering, i_enter);
		update_reduced_costs(p, entering, i_enter);
	}

	// Ported from qdetermine_theta.m
	double determine_theta(const Bu_solution& u, variable_type& leaving,
		size_t& i_leave) const
	{
		// Direction z0
		double theta_z0 = std::numeric_limits<double>::infinity();
		if (u.z0 > 0.0)
			theta_z0 = z0() / u.z0;

		// Direction z_i
		double zi_u_min = std::numeric_limits<double>::infinity();
		size_t zi_u_min_idx = 0;

		// Specialize for the case where u.z0 is negative, as we can save a
		// lot of computation then
		if (u.z0 < 0.0) {
			// Only two elements have the chance to have positive u.z(i).
			// These are u.i_enter (if z_{i_enter} is basic) and u.k.
			if (is_basic_z(u.i_enter)) {
				double u_zi = u.z(u.i_enter);
				if (u_zi > 0.0) {
					double zi_u = z(u.i_enter) / u_zi;
					if (zi_u < zi_u_min) {
						zi_u_min = zi_u;
						zi_u_min_idx = u.i_enter;
					}
				}
			}

			// u.z_k
			assert(is_basic_z(u.k));
			double u_zi = u.z(u.k);
			if (u_zi > 0.0) {
				double zi_u = z(u.k) / u_zi;
				if (zi_u < zi_u_min) {
					zi_u_min = zi_u;
					zi_u_min_idx = u.k;
				}
			}
		} else {
			for (size_t bzi = basis_z.find_first();
				bzi != basis_set::npos; bzi = basis_z.find_next(bzi))
			{
				double u_zi = u.z(bzi);
				if (u_zi > 0.0) {
					double zi_u = z(bzi) / u_zi;
					if (zi_u < zi_u_min) {
						zi_u_min = zi_u;
						zi_u_min_idx = bzi;
					}
				}
			}
		}

		// Direction s_i
		double si_u_min = std::numeric_limits<double>::infinity();
		size_t si_u_min_idx = 0;

		// Specialize for the case where u.z0 is negative, as we can save a
		// lot of computation then
		if (u.z0 < 0.0) {
			// Again only s_{i_enter} and s_k have the chance
			if (is_basic_s(u.i_enter)) {
				double u_si = u.s(u.i_enter);
				if (u_si > 0.0) {
					double si_u = s(u.i_enter) / u_si;
					if (si_u < si_u_min) {
						si_u_min = si_u;
						si_u_min_idx = u.i_enter;
					}
				}
			}
			assert(is_basic_s(u.k));
			double u_si = u.s(u.k);
			if (u_si > 0.0) {
				double si_u = s(u.k) / u_si;
				if (si_u < si_u_min) {
					si_u_min = si_u;
					si_u_min_idx = u.k;
				}
			}
		} else {
			for (size_t bsi = basis_s.find_first();
				bsi != basis_set::npos; bsi = basis_s.find_next(bsi))
			{
				double u_si = u.s(bsi);
				if (u_si > 0.0) {
					double si_u = s(bsi) / u_si;
					if (si_u < si_u_min) {
						si_u_min = si_u;
						si_u_min_idx = bsi;
					}
				}
			}
		}

		if (theta_z0 <= zi_u_min && theta_z0 <= si_u_min) {
			std::cout << "z0 wants to leave the basis, should not happen!  abort."
				<< std::endl;
			assert(0);
		}
		if (zi_u_min <= si_u_min) {
			leaving = z_type;
			i_leave = zi_u_min_idx;
			return zi_u_min;
		} else {
			leaving = s_type;
			i_leave = si_u_min_idx;
			return si_u_min;
		}
	}

	size_t simplex_primal(void) {
		bool update = true;
		size_t count = 0;

		while (update) {
			update = false;

			// Find negative reduced cost variable z_i
			for (size_t i = 0; i < n; ++i) {
				if (is_basic_z(i))
					continue;

				if (redcost_z(i) < -zero_tol) {
					simplex_primal(z_type, i);

					count += 1;
					update = true;
					break;
				}
			}
		}
		return count;
	}

	// Ported from qsimplex_primal.m
	// Return type of leaving variable
	variable_type simplex_primal(variable_type entering, size_t i_enter) {
		if (entering == z_type) {
			// z_{i_enter} is entering the basis
			assert(is_basic_z(i_enter) == false);
		} else if (entering == s_type) {
			// s_{i_enter} is entering the basis
			assert(is_basic_s(i_enter) == false);
		} else {
			// z0 should never leave the basis
			assert(0);
		}

		auto u = (entering == z_type) ? solve_Bu_Azj(i_enter) : solve_Bu_Asj(i_enter);
		variable_type leaving = none;
		size_t i_leave = 0;
		double theta = determine_theta(u, leaving, i_leave);

		update_primal_solution(u, theta, leaving, i_leave, entering, i_enter);

		return leaving;
	}

	void find_entering_variable(variable_type leave_type, size_t i_leave,
		variable_type& enter_type, size_t& i_enter, double& max_w) const
	{
		// Solve pB = e, with e having one element at the leaving variable's
		// basis heading position
		auto p = solve_pB_e(leave_type, i_leave);

		// Determine entering variable
		double max_cw = -std::numeric_limits<double>::infinity();
		max_w = 0.0;
		double min_w = std::numeric_limits<double>::infinity();
		enter_type = none;
		i_enter = 0;
#if 0
		std::cout << "# p.alpha " << p.alpha << "  p.beta " << p.beta << std::endl;
#endif
		for (size_t i = 0; i < n; ++i) {
			// TODO: the most popular update is entering z_i, hence this
			// following block is the hot path
			if (is_basic_z(i) == false) {
				// Non-basic z_i
				double w_zi = p.alpha*D[i] + p.beta*B[i] + p.gamma(i);
				min_w = std::min(w_zi, min_w);
				if (w_zi < 0.0) {
					double rzi = redcost_z(i);
					if ((rzi/w_zi) > max_cw) {
#if 0
						std::cout << "   new max (i " << i << "), rzi "
							<< rzi << "  w_zi " << w_zi << "  p.gamma(i) "
							<< p.gamma(i) << "  rzi/w_zi " << rzi/w_zi
							<< std::endl;
#endif
						max_cw = rzi/w_zi;
						max_w = w_zi;
						enter_type = z_type;
						i_enter = i;
					}
				}
			} else {
				// w_si can be negative only if the following conditions are met:
				//   1. z_i is basic, and
				//   2. s_i is non-basic
				if (is_basic_s(i) == false)
					continue;

				double w_si = p.gamma(i);
				min_w = std::min(w_si, min_w);
				if (w_si < 0.0) {
					double rsi = redcost_s(i);
					if ((rsi/w_si) > max_cw) {
						max_cw = rsi/w_si;
						max_w = w_si;
						enter_type = s_type;
						i_enter = i;
					}
				}
			}
#if 0
			// TODO: this code path can be optimized to only consider negative
			// gamma?
			if (is_basic_s(i) == false) {
				// Non-basic s_i
				double w_si = p.gamma(i);
				min_w = std::min(w_si, min_w);
				if (w_si < 0.0) {
					double rsi = redcost_s(i);
					if ((rsi/w_si) > max_cw) {
						max_cw = rsi/w_si;
						max_w = w_si;
						enter_type = s_type;
						i_enter = i;
					}
				}
			}
#endif
		}

#if 0
		std::cout << "###" << std::endl;
#endif
	}

	// Ported from qsimplex_dual.m
	//
	// vtype: variable type leaving the basis.
	// j: basis-leaving variable index, either z_j (vtype == z_type) or
	//    s_j (vtype == s_type).
	//
	// Return the variable type of the entering variable or none if no
	// variable could be found.
	variable_type simplex_dual(variable_type leave_type, size_t i_leave) {
		assert(i_leave >= 0 && i_leave < n);

		// Assert infeasibility of leaving variable
		double x = (leave_type == z_type) ? z(i_leave) : s(i_leave);
		if (x > 1.0e-10) {
			std::cout << "Forced leaving variable is feasible ("
				<< (leave_type == z_type ? "z_" : "s_") << i_leave << " at value "
				<< x << "), abort." << std::endl;
			assert(0);
		}
#if 0
		std::cout << "DUAL SIMPLEX, leaving " << (leave_type == z_type ? "z_" : "s_")
			<< i_leave << " at " << x << std::endl;
#endif

		double max_w = 0.0;
		variable_type enter_type = none;
		size_t i_enter = 0;
		find_entering_variable(leave_type, i_leave, enter_type, i_enter, max_w);
		if (enter_type == none && std::fabs(xi-xi_max) <= 1.0e-2)
			return none;

		if (enter_type == none) {
			std::cout << "INFEAS: xi " << xi << "  xi_max " << xi_max
				<< "  diff: " << (xi_max-xi) << std::endl;

			dump_problem_instance("infeasible.txt");
			std::cout << "Problem is infeasible!" << std::endl;
			assert(0);
		}
		if (enter_type != z_type) {
			std::cout << "Entering variable is non-basic s_" << i_enter
				<< ", not implemented!" << std::endl;
			assert(0);
		}

		// Solve Bu = a_j and obtain a direction u
		auto u = solve_Bu_Azj(i_enter);
		double theta = x / max_w;

		// Update solution
		update_primal_solution(u, theta, leave_type, i_leave, enter_type, i_enter);
		return enter_type;
	}

	// Ported from qinit.m
	void construct_initial_solution(void) {
		xi_min = B0/D0;
		xi_max = (B0 + std::accumulate(B.cbegin(), B.cend(), 0.0)) /
			(D0 + std::accumulate(D.cbegin(), D.cend(), 0.0));
		xi = xi_min;

		// Maximize A(i)+xi*C(i) to find optimal basis_z
		double i_max_val = -std::numeric_limits<double>::infinity();
		size_t i_max = 0;
		for (size_t i = 0; i < n; ++i) {
			double sc = A[i] + xi*C[i];// + xi*(B[i]/D[i]);	// FIXME
			if (sc <= i_max_val)
				continue;

			i_max_val = sc;
			i_max = i;
		}
		basis_z[i_max] = true;	// z_{i_max} is basic
		k_shared = i_max;	// z_k and s_k are in basis

		for (size_t i = 0; i < n; ++i)
			basis_s[i] = true;

		D0_basis_sum = D0 + D[i_max];
		B0_basis_sum = B0 + B[i_max];
		A_basis_sum = A[i_max];
		C_basis_sum = C[i_max];

		// xi-independent z0, z, s
		auto ms = solve_2by2(D0, D[i_max], B0, B[i_max], 1.0, 0.0);
#ifdef CHECK_PRIMAL
		z0_noxi = ms.first;	// z_0
#endif
		z0_noxi2 = ms.first;	// z_0
#ifdef CHECK_PRIMAL
		z_noxi[i_max] = ms.second;	// z_{i_max}
		for (size_t i = 0; i < n; ++i)
			s_noxi[i] = z0_noxi;
		s_noxi[i_max] = ms.first - ms.second;
#endif

		z_noxi2 = basic_noxi_vector(z_type, z0_noxi2);
		z_noxi2.fix_value(i_max, ms.second);

		s_noxi2 = basic_noxi_densevector(n, s_type, z0_noxi2);
		s_noxi2.fix_value(i_max, ms.first - ms.second);
#ifdef CHECK_PRIMAL
		for (size_t bzi = basis_z.find_first();
			bzi != basis_set::npos; bzi = basis_z.find_next(bzi))
		{
			assert(std::fabs(z_noxi[bzi] - z_noxi2(bzi)) <= 1.0e-14);
		}
		for (size_t bsi = basis_s.find_first();
			bsi != basis_set::npos; bsi = basis_s.find_next(bsi))
		{
			assert(std::fabs(s_noxi[bsi] - s_noxi2(bsi)) <= 1.0e-14);
		}
#endif

		// xi-dependent z0, z, s
		ms = solve_2by2(D0, D[i_max], B0, B[i_max], 0.0, 1.0);
#ifdef CHECK_PRIMAL
		z0_xi = ms.first;	// z_0
#endif
		z0_xi2 = ms.first;
#ifdef CHECK_PRIMAL
		z_xi[i_max] = ms.second;	// z_{i_max}, k == i_max
		for (size_t i = 0; i < n; ++i)
			s_xi[i] = z0_xi;
		s_xi[i_max] = ms.first - ms.second;	// z_0 - z_{i_max}
#endif

		// z_xi[i_max] = z_{i_max} = z0 - (z0 - z_{i_max})
		z_xi2 = Bu_solution(this, z0_xi2, z0_xi2 - ms.second, i_max);
		s_xi2 = Bu_solution(this, z0_xi2, z0_xi2 - ms.second, i_max);
#ifdef CHECK_PRIMAL
		for (size_t i = 0; i < n; ++i) {
			assert(std::fabs(z_xi[i] - z_xi2.z(i)) <= 1.0e-14);
			assert(std::fabs(s_xi[i] - s_xi2.s(i)) <= 1.0e-14);
		}

		// TODO
		std::cout << "initial solution check" << std::endl;
		check_primal();
#endif

		// Reduced costs, xi-independent z, s
		ms = solve_2by2(D0, B0, D[i_max], B[i_max], 0.0, -A[i_max]);
		double alpha = ms.first;
		double beta = ms.second;
		redcost_z0_noxi = alpha*D0 + beta*B0;
		for (size_t i = 0; i < n; ++i) {
			redcost_z_noxi[i] = alpha*D[i] + beta*B[i];
			redcost_s_noxi[i] = 0;
		}
#ifdef REDCOST_TODO
		redcost_z_noxi2.add_tau(alpha, beta);
#endif

		// Reduced costs, xi-dependent z, s
		ms = solve_2by2(D0, B0, D[i_max], B[i_max], -C0, -C[i_max]);
		alpha = ms.first;
		beta = ms.second;
#if 0
		redcost_z0_xi = alpha*D0 + beta*B0;
#endif
		for (size_t i = 0; i < n; ++i) {
			redcost_z_xi[i] = alpha*D[i] + beta*B[i];
			redcost_s_xi[i] = 0;
		}
#ifdef REDCOST_TODO
		redcost_z_xi2.add_tau(alpha, beta);
#endif
	}

	// Ported from qcompute_xi_range.m
	//
	// This is the only place where the reduced costs are used (they are
	// updated in update_reduced_costs and initialized in
	// construct_initial_solution).
	std::pair<double, double> compute_xi_range(limiting_type& btype,
		variable_type& vtype, size_t& j) const
	{
		// From dual feasibility condition: dN + xi cN >= 0
		double alpha_1 = -std::numeric_limits<double>::infinity();
		double beta_1 = std::numeric_limits<double>::infinity();
		size_t beta_1_idx = 0;
		variable_type beta_1_type = none;	// 1: z_i, 2: s_i

		// From primal feasibility condition
		double alpha_2 = -std::numeric_limits<double>::infinity();
		double beta_2 = std::numeric_limits<double>::infinity();
		size_t beta_2_idx = 0;
		variable_type beta_2_type = none;	// 1: z_i, 2: s_i

		for (size_t i = 0; i < n; ++i) {
			// z_i
			if (is_basic_z(i)) {
				// Basic z_i
				double p = z_xi2.z(i);
#ifdef CHECK_PRIMAL
				double p1 = z_xi[i];
				assert(std::fabs(p1 - p) <= 1.0e-14);
#endif
#ifdef ALPHA
				if (p > zero_tol) {
					double q = z_noxi2(i);
					alpha_2 = std::max(alpha_2, -q/p);
				} else
#endif
				if (p < -zero_tol) {
					double q = z_noxi2(i);
#ifdef CHECK_PRIMAL
					double q1 = z_noxi[i];
					assert(std::fabs(q1 - q) <= 1.0e-14);
#endif
					if (-q/p < beta_2) {
						beta_2 = -q/p;
						beta_2_idx = i;
						beta_2_type = z_type;	// z_i
					}
				}
			} else {
				// Non-basic z_i
				double cN = cost_z_xi(i) - redcost_z_xi[i];
#ifdef ALPHA
				if (cN > zero_tol) {
					double dN = cost_z_noxi(i) - redcost_z_noxi[i];
					alpha_1 = std::max(alpha_1, -dN/cN);
				} else
#endif
				if (cN < -zero_tol) {
					double dN = cost_z_noxi(i) - redcost_z_noxi[i];
					if (-dN/cN < beta_1) {
						beta_1 = -dN/cN;
						beta_1_idx = i;
						beta_1_type = z_type;	// z_i
					}
				}
			}

			// s_i
			if (is_basic_s(i)) {
				// Basic s_i
				double p = s_xi2.s(i);
#ifdef CHECK_PRIMAL
				double p1 = s_xi[i];
				assert(std::fabs(p1 - p) <= 1.0e-14);
#endif
#ifdef ALPHA
				if (p > zero_tol) {
					double q = s_noxi2(i);
					alpha_2 = std::max(alpha_2, -q/p);
				} else
#endif
				if (p < -zero_tol) {
					double q = s_noxi2(i);
#ifdef CHECK_PRIMAL
					double q1 = s_noxi[i];
					assert(std::fabs(q1 - q) <= 1.0e-14);
#endif
					if (-q/p < beta_2) {
						beta_2 = -q/p;
						beta_2_idx = i;
						beta_2_type = s_type;	// s_i
					}
				}
			} else {
				// Non-basic s_i
				double cN = -redcost_s_xi[i];
#ifdef ALPHA
				if (cN > zero_tol) {
					double dN = -redcost_s_noxi[i];
					alpha_1 = std::max(alpha_1, -dN/cN);
				} else
#endif
				if (cN < -zero_tol) {
					double dN = -redcost_s_noxi[i];
					if (-dN/cN < beta_1) {
						beta_1 = -dN/cN;
						beta_1_idx = i;
						beta_1_type = s_type;	// s_i
					}
				}
			}
		}
		assert(beta_1_type != none || beta_2_type != none);

		// Determine limiting bound
		if (beta_1 < beta_2) {
			btype = dual_feasibility;
			j = beta_1_idx;
			vtype = beta_1_type;
		} else {
			btype = primal_feasibility;
			j = beta_2_idx;
			vtype = beta_2_type;
		}

		double alpha = std::max(alpha_1, alpha_2);
		double beta = std::min(beta_1, beta_2);
		if (beta < (xi_min-1.0e-5) || beta > (xi_max+1.0e-5)) {
			std::cout << "btype " << btype << "  j " << j << "  vtype " << vtype
				<< "  beta " << beta << "  xi_min " << xi_min
				<< "  xi_max " << xi_max << std::endl;
			dump_problem_instance("problem.txt");
			assert(0);
		}

		return std::pair<double, double>(alpha, beta);
	}

	void extract_primal_solution(std::vector<double>& x_opt,
		double best_xi) const
	{
		assert(best_xi >= xi_min && best_xi <= xi_max);
		assert(x_opt.size() == n);

		double opt_z0 = z0(best_xi);
		for (size_t i = 0; i < n; ++i) {
			x_opt[i] = is_basic_z(i) ? (z(i,best_xi) / opt_z0) : 0.0;
			if (x_opt[i] < -1.0e-9 || x_opt[i] > 1.0+1.0e-9) {
				std::cout << "x_opt[" << i << "] " << x_opt[i] << std::endl;
				std::cout << "best_xi " << best_xi
					<< "  xi_min " << xi_min << "  xi_max " << xi_max << std::endl;
				std::cout << "z(i,best_xi) " << z(i,best_xi) << std::endl;
				std::cout << "opt_z0 " << opt_z0 << std::endl;
			}
			assert(x_opt[i] >= -1.0e-9 && x_opt[i] <= (1.0+1.0e-9));
		}
	}

	void update_le_stat(solve_statistics& stat,
		variable_type leave_type, variable_type enter_type) const
	{
		if (leave_type == z_type && enter_type == z_type) {
			stat.lzez += 1;
		} else if (leave_type == z_type && enter_type == s_type) {
			stat.lzes += 1;
		} else if (leave_type == s_type && enter_type == z_type) {
			stat.lsez += 1;
		} else if (leave_type == s_type && enter_type == s_type) {
			stat.lses += 1;
		}
	}

public:
	problem(double A0, const vector& A, double B0, const vector& B,
		double C0, const vector& C, double D0, const vector& D)
		: n(A.size()), A0(A0), A(A), B0(B0), B(B), C0(C0), C(C), D0(D0), D(D),
			basis_z(n), basis_s(n),
#ifdef CHECK_PRIMAL
			z_xi(n, 0), z_noxi(n, 0), s_xi(n, 0), s_noxi(n, 0),
#endif
			redcost_z_xi(n, 0), redcost_z_noxi(n, 0),
#ifdef REDCOST_TODO
			redcost_z_xi2(this, n, z_type), redcost_z_noxi2(this, n, z_type),
#endif
			redcost_s_xi(n, 0), redcost_s_noxi(n, 0)
#ifdef REDCOST_TODO
			redcost_s_xi2(this, n, s_type), redcost_s_noxi2(this, n, s_type)
#endif
	{
		assert(n == B.size());
		assert(n == C.size());
		assert(n == D.size());

		construct_initial_solution();
	}

	bool check_instance_ok(double tol = 1.0e-10) const {
		// Verify required feasibility assumptions
		// 1. B0 + sum_i B_i lambda_i > 0 for all lambda
		if (B0 <= tol)
			return false;	// B_i >= 0

		// 2. D0 + sum_i D_i lambda_i > 0 for all lambda
		double dsum = D0 + std::accumulate(D.cbegin(), D.cend(), 0.0);
		if (dsum <= tol)
			return false;

		return true;
	}

	static double fractional_rate(const std::vector<double>& sol,
		double frac_tol = 1.0e-3)
	{
		size_t fractional_count = 0;
		for (size_t i = 0; i < sol.size(); ++i) {
			if (sol[i] >= frac_tol && sol[i] <= (1.0-frac_tol))
				fractional_count += 1;
		}
		return static_cast<double>(fractional_count) /
			static_cast<double>(sol.size());
	}

	// Compute P(xi) objective
	double objective(const std::vector<double>& x) const {
		assert(x.size() == n);

		double n1 = A0;
		double d1 = B0;
		double n2 = C0;
		double d2 = D0;
		for (size_t i = 0; i < n; ++i) {
			n1 += A[i] * x[i];
			d1 += B[i] * x[i];
			n2 += C[i] * x[i];
			d2 += D[i] * x[i];
		}
		return n1/d1 + n2/d2;
	}

	// Return globally optimal objective value and solution, possibly
	// fractional.
	solve_statistics solve(std::vector<double>& x_opt) {
		// Keep track of runtime
		timer solve_timer;
		solve_timer.start();

		x_opt.resize(n);

		solve_statistics stat;
		stat.n = n;
		stat.primal_simplex_init_count = simplex_primal();

		stat.f_xi_opt = -std::numeric_limits<double>::infinity();
		stat.xi_opt = std::numeric_limits<double>::signaling_NaN();
		for (int iter = 1; true; ++iter) {
			assert(xi >= xi_min && xi <= xi_max);
#if 0
			double obj_lp = pobj_lp();
			double obj = -obj_lp / xi;
			double inf_primal = pinf();
			std::cout << "[lfp] iter " << iter << "  xi " << xi
				<< "  obj_lp " << obj_lp << "  obj " << obj
				<< "  pinf " << inf_primal << std::endl;
#endif

			limiting_type btype;
			variable_type vtype;
			size_t active_j;
			auto xi_range = compute_xi_range(btype, vtype, active_j);
			assert(xi_range.second >= (xi_min-1.0e-8));
			assert(xi_range.second <= (xi_max+1.0e-8));
#if 0
			std::cout << "   xi_range: [" << xi_range.first << " ; "
				<< xi_range.second << "]" << std::endl;
#endif

			// Compute quadratic objective in xi
			auto obj_t = compute_quadratic_objective();
			double f_max = std::numeric_limits<double>::signaling_NaN();
			double xi_mode = compute_objective_maximizer(obj_t,
				xi, xi_range.second, f_max);
			if (xi_mode != xi_range.second && f_max > stat.f_xi_opt) {
				// Not at the right border, save solution
#if 0
				std::cout << "iteration " << iter << " at xi " << xi_mode
					<< " in [" << xi_min << " ; " << xi_max << "]  obj "
					<< f_max << std::endl;
#endif
				stat.f_xi_opt = f_max;
				stat.xi_opt = xi_mode;
				extract_primal_solution(x_opt, stat.xi_opt);
				stat.primal_extract_count += 1;
				stat.opt_time_ms = solve_timer.elapsed_ms();
			}

			xi = xi_range.second;
			if (xi >= (xi_max - 1.0e-10))
				break;

			if (btype == dual_feasibility) {
				// active_j is entering the basis
				variable_type leave_type = simplex_primal(vtype, active_j);
				stat.primal_simplex_count += 1;

				update_le_stat(stat, leave_type, vtype /* enter_type */);
			} else if (btype == primal_feasibility) {
				// active_j is leaving the basis
				variable_type enter_type = simplex_dual(vtype, active_j);
				if (enter_type == none) {
					// xi \approx xi_max, no entering variable found
					break;
				}

				stat.dual_simplex_count += 1;
				update_le_stat(stat, vtype /* leave_type */, enter_type);
			} else {
				assert(0);
			}
		}

		// Verify if an optimal solution has been recovered yet, if not it is
		// at the extreme right side of xi = xi_max, hence force a solution
		// recovery
		if (stat.f_xi_opt <= -std::numeric_limits<double>::infinity()) {
			auto obj_t = compute_quadratic_objective();
			double f_max = std::get<0>(obj_t)/xi_max
				+ std::get<1>(obj_t) + xi_max*std::get<2>(obj_t);

			stat.f_xi_opt = f_max;
			stat.xi_opt = xi_max;
			extract_primal_solution(x_opt, xi_max);
			stat.primal_extract_count += 1;
			stat.opt_time_ms = solve_timer.elapsed_ms();
		}

		stat.fract_opt = fractional_rate(x_opt);
		stat.total_time_ms = solve_timer.elapsed_ms();

#if 0
		std::cout << stat;
#endif

		return stat;
	}

	// Output textual representation of the current problem status
	friend std::ostream& operator<<(std::ostream& os, const problem& prob) {
		std::ios_base::fmtflags os_flags(os.flags());
		os << "Problem (n=" << prob.n << ")" << std::endl;
		os << "   basis_z (" << prob.basis_z.count() << " elements):";
		for (size_t i = 0; i < prob.n; ++i) {
			if (prob.is_basic_z(i) == false)
				continue;
			os << " z_" << i << ",";
		}
		os << std::endl;
		os << "   basis_s (" << prob.basis_s.count() << " elements):";
		for (size_t i = 0; i < prob.n; ++i) {
			if (prob.is_basic_s(i) == false)
				continue;
			os << " s_" << i << ",";
		}
		os << std::endl;

		os << std::fixed << std::scientific << std::setprecision(8);
			//<< std::setprecision(std::numeric_limits<double>::digits10);
		os << "   xi " << prob.xi << " in [" << prob.xi_min
			<< " ; " << prob.xi_max << "]" << std::endl;

		std::cout << "Printing basic variables" << std::endl;

		// z0
		os << "   z0 " << prob.z0();
		os << "   z0_xi " << prob.z0_xi2;
		os << "   z0_noxi " << prob.z0_noxi2;
		os << std::endl;

		// basic z_i
		for (size_t i = 0; i < prob.n; ++i) {
			if (prob.is_basic_z(i) == false)
				continue;

			os << "   z_" << i << " " << prob.z(i)
				<< "  z_" << i << "_xi " << prob.z_xi2.z(i)
				<< "  z_" << i << "_noxi " << prob.z_noxi2(i)
				<< std::endl;
		}
		// basic s_i
		for (size_t i = 0; i < prob.n; ++i) {
			if (prob.is_basic_s(i) == false)
				continue;

			os << "   s_" << i << " " << prob.s(i)
				<< "  s_" << i << "_xi " << prob.s_xi2.s(i)
				<< "  s_" << i << "_noxi " << prob.s_noxi2(i)
				<< std::endl;
		}

		std::cout << "Printing reduced costs (basic and non-basic)" << std::endl;
		os << "   B ~z0 " << prob.redcost_z0()
			<< "  cost " << prob.cost_z0() << " - xi * "
			<< prob.redcost_z0_xi << " - " << prob.redcost_z0_noxi
			<< std::endl;
		for (size_t i = 0; i < prob.n; ++i) {
			os << "   " << (prob.is_basic_z(i) ? "B" : "N");
			os << " ~z_" << i << " " << prob.redcost_z(i)
				<< "  cost " << prob.cost_z(i) << " - xi * "
				<< prob.redcost_z_xi[i] << " - " << prob.redcost_z_noxi[i]
				<< std::endl;
		}
		for (size_t i = 0; i < prob.n; ++i) {
			os << "   " << (prob.is_basic_s(i) ? "B" : "N");
			os << " ~s_" << i << " " << prob.redcost_s(i)
				<< "  cost " << prob.cost_s(i) << " - xi * "
				<< prob.redcost_s_xi[i] << " - " << prob.redcost_s_noxi[i]
				<< std::endl;
		}

		os.flags(os_flags);

		return os;
	}

	// Write a text serialization of the instance to the given file path
	void dump_problem_instance(const std::string& path) const {
		std::ofstream ofs(path);
		ofs << n << std::endl;
		ofs << std::fixed << std::scientific
			<< std::setprecision(std::numeric_limits<double>::digits10);

		// A0, A
		ofs << A0 << std::endl;
		for (size_t i = 0; i < n; ++i) {
			if (i > 0)
				ofs << " ";
			ofs << A[i];
		}
		ofs << std::endl;

		// B0, B
		ofs << B0 << std::endl;
		for (size_t i = 0; i < n; ++i) {
			if (i > 0)
				ofs << " ";
			ofs << B[i];
		}
		ofs << std::endl;

		// C0, C
		ofs << C0 << std::endl;
		for (size_t i = 0; i < n; ++i) {
			if (i > 0)
				ofs << " ";
			ofs << C[i];
		}
		ofs << std::endl;

		// D0, D
		ofs << D0 << std::endl;
		for (size_t i = 0; i < n; ++i) {
			if (i > 0)
				ofs << " ";
			ofs << D[i];
		}
		ofs << std::endl;

		ofs.close();
	}

};

}

#endif

