
#include <iostream>
#include <iomanip>
#include <fstream>
#include <cmath>
#include <random>

#include "lfp.h"
#include "timing.h"
#include "utility.h"

typedef lfp::problem::vector vector;

void bench_random(size_t n, size_t reps, unsigned long seed = 0) {
	std::random_device gseed;
	std::mt19937 eng(seed == 0 ? gseed() : seed);
	std::uniform_real_distribution<double> runi;

	double A0 = 0.0;
	vector A(n, 0.0);
	double B0 = 0.0;
	vector B(n, 0.0);
	double C0 = 0.0;
	vector C(n, 0.0);
	double D0 = n;
	vector D(n, 0.0);

	mean_variance_accumulator<double> runtime;
	mean_variance_accumulator<double> fract;
	for (size_t ri = 0; ri < reps; ++ri) {
		A0 = 0.5 * static_cast<double>(n) * runi(eng);
		for (size_t i = 0; i < n; ++i) {
			double p1 = runi(eng);
			double p2 = runi(eng);
			double cs = runi(eng)/(p1+p2);
			p1 *= cs;
			p2 *= cs;

			A[i] = p1;
			B0 += p1;
			B[i] = 1.0 - p1;
			C0 += p2;
			C[i] = -p2;
			D[i] = p2-1.0;
		}
		lfp::problem prob(A0, A, B0, B, C0, C, D0, D);

		std::vector<double> sol;

		timer t;
		t.start();
		auto stat = prob.solve(sol);
		t.stop();

		runtime.observe(t.elapsed_ms(), 1.0);
		fract.observe(stat.fract_opt, 1.0);
	}

	std::cout << std::fixed << std::setprecision(std::numeric_limits<double>::digits10);

	// Standard error
	double stddev = std::sqrt(runtime.sample_variance());

	// n   mean-runtime(ms)   std-runtime(ms)   fract-var-%
	std::cout << n << " " << runtime.mean() << " " << stddev
		<< " " << (fract.mean()*100.0) << std::endl;
}

int main(int argc, char* argv[]) {
	size_t reps = 10;

	for (size_t n = 1000; n <= 50000; n += 1000) {
		bench_random(n, reps);
	}
}

