
#ifndef UTILITY_H
#define UTILITY_H

#include <algorithm>
#include <vector>
#include <limits>
#include <iostream>
#include <unordered_map>
#include <cassert>

// Online weighted running mean and variance updates
//
// Reference: One-pass algorithm WV2 in
// [West1979] D.H.D. West, "Updating Mean and Variance Estimates: An
//            Improved Method", Comm. of the ACM, Vol. 22, Issue 9,
//            532--535, 1979.
template<typename TFloat>
class mean_variance_accumulator {
private:
	TFloat sumw;
	TFloat wmean;
	TFloat t;
	size_t n;

	TFloat vmin;
	TFloat vmax;

public:
	mean_variance_accumulator() : sumw(0), wmean(0), t(0), n(0),
		vmin(std::numeric_limits<TFloat>::infinity()),
		vmax(-std::numeric_limits<TFloat>::infinity()) {
	}

	// Add a weighted observation x_i with weight w_i
	void observe(TFloat value, TFloat weight) {
		TFloat q = value - wmean;
		TFloat temp_sumw = sumw + weight;
		TFloat r = q * weight / temp_sumw;
		wmean += r;
		t += r * sumw * q;
		sumw = temp_sumw;
		n += 1;

		vmin = std::min(vmin, value);
		vmax = std::max(vmax, value);
	}

	// Return the weighted sample mean,
	// xbar = (sum_i w_i x_i) / (sum_i w_i)
	TFloat mean(void) const {
		return wmean;
	}

	// Return the smallest value observed
	TFloat min(void) const {
		return vmin;
	}
	// Return the largest value observed
	TFloat max(void) const {
		return vmax;
	}

	// Return the weighted population variance,
	//    S^2 = (sum_i w_i (x_i - xbar)^2) / (sum_i w_i)
	TFloat population_variance(void) const {
		assert(n >= 1);
		return t / sumw;
	}

	// Return the weighted sample variance,
	//    S^2 = (sum_i w_i (x_i - xbar)^2) / (((n-1)/n) sum_i w_i),
	// where n is the number of observations.
	//
	// For the special case of w_i = 1 for all i we have,
	//    S^2 = (sum_i (x_i - xbar)^2) / (n-1),
	// the ordinary sample variance.
	TFloat sample_variance(void) const {
		assert(n > 1);
		TFloat fac = static_cast<TFloat>(n) /
			(static_cast<TFloat>(n-1) * sumw);

		return t * fac;
	}
};

#endif

