
#ifndef VOC_H
#define VOC_H

#include <algorithm>
#include <numeric>
#include <set>
#include <functional>
#include <random>
#include <tuple>
#include <utility>
#include <cstdint>
#include <cassert>

#include "lfp.h"
#include "pairpartition.h"

namespace voc {

/* Optimal prediction function under the intersection-over-union score, given
 * calibrated marginal posterior beliefs.
 *
 * Author: Sebastian Nowozin <Sebastian.Nowozin@microsoft.com>
 * Date: 16th October 2013.
 */
template<typename TFloat>
class prediction_function {
public:
	typedef std::uint8_t class_t;
	typedef std::function<TFloat(size_t i, size_t c)> pfunc_t;
	typedef std::function<class_t(size_t i)> initfunc_t;

private:
	pfunc_t P;

	size_t class_count;	// Class count
	size_t n;	// Number of instances

	std::vector<class_t> y;

	void check_posterior(void) const {
		for (size_t i = 0; i < n; ++i) {
			double psum = 0.0;
			for (size_t c = 0; c < class_count; ++c)
				psum += P(i,c);

			assert(std::fabs(psum - 1.0) <= 1.0e-7);
		}
	}

	// Relabel within an arbitrary subset W
	size_t relabel_c1_c2(const std::set<size_t>& W, class_t c1, class_t c2) {
		double A0 = 0.0;
		double B0 = 0.0;
		double C0 = 0.0;
		double D0 = 0.0;

		size_t n_c1c2 = 0;
		for (auto wi = W.cbegin(); wi != W.cend(); ++wi) {
			assert(*wi < n);
			if (y[*wi] == c1 || y[*wi] == c2)
				n_c1c2 += 1;
		}
		if (n_c1c2 == 0)
			return 0;

		std::vector<double> A(n_c1c2, 0.0);
		std::vector<double> B(n_c1c2, 0.0);
		std::vector<double> C(n_c1c2, 0.0);
		std::vector<double> D(n_c1c2, 0.0);

		// Current solution
		std::vector<double> x_cur(n_c1c2, 0.0);

		// TODO: can do more efficient sums here by keeping the total
		size_t j = 0;
		for (size_t i = 0; i < n; ++i) {
			bool in_W = (W.find(i) != W.end()) && (y[i] == c1 || y[i] == c2);

			B0 += P(i,c1);
			if (in_W) {
				// in W
				assert(j < n_c1c2);
				A[j] = P(i,c1);
				B[j] = 1.0 - P(i,c1);
				C[j] = -P(i,c2);
				D[j] = P(i,c2) - 1.0;

				C0 += P(i,c2);
				D0 += 1.0;

				x_cur[j] = (y[i] == c1) ? 1.0 : 0.0;

				j += 1;
			} else {
				// not in W
				A0 += (y[i] == c1) ? P(i,c1) : 0.0;
				B0 += (y[i] == c1) ? (1.0-P(i,c1)) : 0.0;
				C0 += (y[i] == c2) ? P(i,c2) : 0.0;
				D0 += P(i,c2);
				D0 += (y[i] == c2) ? (1.0-P(i,c2)) : 0.0;
			}
		}
		assert(j == n_c1c2);

		// Solve
		lfp::problem prob(A0, A, B0, B, C0, C, D0, D);
		assert(prob.check_instance_ok());
		double obj_cur = prob.objective(x_cur);

		std::vector<double> x_opt(n_c1c2, 0.0);
		auto stat = prob.solve(x_opt);

#if 0
		std::cout << "obj_cur " << obj_cur << "  f_xi_opt " << stat.f_xi_opt << std::endl;
#endif
		if (stat.f_xi_opt <= (obj_cur - 1.0e-4)) {
			std::cout << "DUMPING PROBLEM to file 'crash1.txt'." << std::endl;
			prob.dump_problem_instance("crash1.txt");
		}
		assert(stat.f_xi_opt >= (obj_cur - 1.0e-4));

		// Decode solution
		j = 0;
		size_t fractional_count = 0;
		for (auto wi = W.cbegin(); wi != W.cend(); ++wi) {
			assert(*wi < n);
			if (y[*wi] != c1 && y[*wi] != c2)
				continue;

			assert(j < x_opt.size());

			if (x_opt[j] >= 1.0e-3 && x_opt[j] <= (1.0-1.0e-3))
				fractional_count += 1;

			y[*wi] = (x_opt[j] >= 0.5) ? c1 : c2;
			j += 1;
		}

		return fractional_count;
	}

	// Return absolute number of fractional solutions
	size_t relabel_c1_c2(class_t c1, class_t c2) {
		double B0 = 0.0;
		double C0 = 0.0;
		double D0 = 0.0;

		// Compute size of task and B0/C0/D0 constants
		size_t n_c1c2 = 0;
		for (size_t i = 0; i < n; ++i) {
			B0 += P(i,c1);
			D0 += P(i,c2);

			if (y[i] == c1 || y[i] == c2) {
				C0 += P(i,c2);
				D0 += 1.0 - P(i,c2);
				n_c1c2 += 1;
			}
		}

		// Nothing to do
		if (n_c1c2 == 0)
			return 0;

		std::vector<double> A(n_c1c2, 0.0);
		std::vector<double> B(n_c1c2, 0.0);
		std::vector<double> C(n_c1c2, 0.0);
		std::vector<double> D(n_c1c2, 0.0);
		size_t j = 0;
		for (size_t i = 0; i < n; ++i) {
			if (y[i] != c1 && y[i] != c2)
				continue;

			// In set W
			A[j] = P(i,c1);
			B[j] = 1.0 - P(i,c1);
			C[j] = -P(i,c2);
			D[j] = P(i,c2) - 1.0;
			j += 1;
		}
		assert(j == n_c1c2);

		// Solve ratio of two linear fractional functions problem
		lfp::problem prob(0.0 /* A0 */, A, B0, B, C0, C, D0, D);
		assert(prob.check_instance_ok());
		std::vector<double> x_opt(n_c1c2, 0.0);
		prob.solve(x_opt);

		// Decode solution
		j = 0;
		size_t fractional_count = 0;
		for (size_t i = 0; i < n; ++i) {
			if (y[i] != c1 && y[i] != c2)
				continue;

			if (x_opt[j] >= 1.0e-3 && x_opt[j] <= (1.0-1.0e-3))
				fractional_count += 1;

			y[i] = (x_opt[j] >= 0.5) ? c1 : c2;
			j += 1;
		}
		return fractional_count;
	}

	void sample_vector(std::vector<class_t>& y, std::mt19937& eng) const {
		std::uniform_real_distribution<double> runi;
		y.resize(n);

		for (size_t i = 0; i < n; ++i) {
			double rval = runi(eng);
			double csum = 0.0;
			class_t c = 0;
			for ( ; c < class_count; ++c) {
				csum += P(i,c);
				if (rval <= csum)
					break;
			}
			y[i] = c;
		}
	}

public:
	enum init_type {
		map_assignment = 0,
		random_assignment = 1,
		inverseweighted_assignment = 2,
	};

	// Standard float array access:
	// P_base[i*class_count_in + c] is posterior probability for class c at
	//    sample i.
	prediction_function(const TFloat* P_base, size_t class_count_in, size_t n_in)
		: P([P_base, class_count_in, n_in](size_t i, size_t c) -> double
			{
				assert(i < n_in && c < class_count_in);
				const TFloat* P_pix = P_base + i*class_count_in;
				return P_pix[c];
			}),
			class_count(class_count_in), n(n_in), y(n_in, 0)
	{
		assert(class_count >= 2);
		assert(n > 0);
		check_posterior();
	}

	// Arbitrary access mechanism
	prediction_function(pfunc_t P, size_t class_count, size_t n)
		: P(P), class_count(class_count), n(n), y(n, 0)
	{
		assert(class_count >= 2);
		assert(n > 0);
		check_posterior();
	}

	const std::vector<class_t>& get_solution(void) const {
		return y;
	}

	// Return the intersection over union score for a discrete labeling
	static std::pair<size_t, size_t> compute_iou_c(const std::vector<class_t>& z,
		const std::vector<class_t>& y, class_t c)
	{
		assert(z.size() == y.size());

		size_t u_intersection = 0;
		size_t u_union = 0;

		for (size_t i = 0; i < z.size(); ++i) {
			u_intersection += (z[i] == c && y[i] == c) ? 1 : 0;
			u_union += (z[i] == c || y[i] == c) ? 1 : 0;
		}

		return std::pair<size_t, size_t>(u_intersection, u_union);
	}

	// Return (E[S_c/T_c], E[S_c], E[T_c])
	std::tuple<double, double, double> compute_iou_c_montecarlo(class_t c,
		size_t reps, unsigned long seed = 0) const
	{
		std::random_device gseed;
		std::mt19937 eng(seed == 0 ? gseed() : seed);

		std::vector<class_t> y(n, class_t(0));
		std::vector<class_t> z(n, class_t(0));

		double EST = 0.0;
		double ES = 0.0;
		double ET = 0.0;
		for (size_t ri = 0; ri < reps; ++ri) {
			sample_vector(y, eng);
			sample_vector(z, eng);

			auto iou = compute_iou_c(z, y, c);
			EST += static_cast<double>(iou.first) /
				static_cast<double>(iou.second);
			ES += static_cast<double>(iou.first);
			ET += static_cast<double>(iou.second);
		}
		EST /= static_cast<double>(reps);
		ES /= static_cast<double>(reps);
		ET /= static_cast<double>(reps);

		return std::tuple<double, double, double>(EST, ES, ET);
	}

	// Return the expected intersection over union score for a probability
	// distribution, approximation from the paper
	std::pair<double, double> compute_objective_c(class_t c) const {
		double A_c = 0.0;
		double D_c = 0.0;

		for (size_t i = 0; i < n; ++i) {
			if (y[i] == c) {
				A_c += P(i,c);
				D_c += 1.0;
			} else {
				D_c += P(i,c);
			}
		}
		return std::pair<double, double>(A_c, D_c);
	}

	double compute_objective(void) const {
		double obj = 0.0;

		for (class_t c = 0; c < class_count; ++c) {
			std::pair<double, double> AD = compute_objective_c(c);
			obj += AD.first / AD.second;
		}
		obj /= static_cast<double>(class_count);

		return obj;
	}

	std::vector<double> compute_objective_perclass(void) const {
		std::vector<double> obj(class_count, 0.0);

		for (class_t c = 0; c < class_count; ++c) {
			std::pair<double, double> AD = compute_objective_c(c);
			obj[c] = AD.first / AD.second;
		}

		return obj;
	}

	// Optimized version of Tarlow's greedy algorithm with single site flips.
	// Complexity is O(N K).
	//
	// Return the number of changed variables.
	size_t greedy_sweep(void) {
		// Random order
		std::vector<size_t> sites(n);
		std::iota(sites.begin(), sites.end(), 0);
		std::random_shuffle(sites.begin(), sites.end());

		// Precompute sufficient statistics needed
		std::vector<double> alpha(class_count, 0.0);
		std::vector<double> beta(class_count, 0.0);
		for (size_t i = 0; i < n; ++i) {
			size_t y_i = y[i];
			alpha[y_i] += P(i,y_i);
			beta[y_i] += (1.0 - P(i,y_i));
			for (class_t c = 0; c < class_count; ++c)
				beta[c] += P(i,c);
		}

		// Compute current objective
		double obj = 0.0;
		for (class_t c = 0; c < class_count; ++c)
			obj += alpha[c] / beta[c];
		obj /= static_cast<double>(class_count);
		//double obj_sweep_start = obj;

		// Perform sweep in random order
		size_t j;
		class_t k1;
		class_t k2;
		size_t changed = 0;
		for (size_t j_o = 0; j_o < n; ++j_o) {
			j = sites[j_o];
			k1 = y[j];	// current label

			// Default: no improvement by remaining with k1
			double delta_f_best = 0.0;
			class_t k_best = k1;

			for (k2 = 0; k2 < class_count; ++k2) {
				if (k1 == k2)
					continue;

				// Compute changed statistics
				double alpha_k2_new = alpha[k2] + P(j,k2);
				double beta_k2_new = beta[k2] + 1.0 - P(j,k2);
				double alpha_k1_new = alpha[k1] - P(j,k1);
				double beta_k1_new = beta[k1] - 1.0 + P(j,k1);

				// Compute improvement
				double delta_f = alpha_k2_new / beta_k2_new
					+ alpha_k1_new / beta_k1_new
					- alpha[k2] / beta[k2]
					- alpha[k1] / beta[k1];
				delta_f /= static_cast<double>(class_count);

				if (delta_f <= delta_f_best)
					continue;

				// Found potential improvement
				delta_f_best = delta_f;
				k_best = k2;
			}

			if (delta_f_best <= 0.0)
				continue;

			// Update solution and statistics
			y[j] = k_best;
			alpha[k_best] += P(j,k_best);
			beta[k_best] += 1.0 - P(j,k_best);
			alpha[k1] -= P(j,k1);
			beta[k1] -= 1.0 - P(j,k1);
			obj += delta_f_best;

			changed += 1;
		}
#if 0
		std::cout << "sweep improved objective from " << obj_sweep_start
			<< " to " << obj << std::endl;
#endif

		return changed;
	}

	// Try to relabel all sites
	void relabel_sweep(void) {
		std::vector<class_t> labels(class_count, 0);
		std::iota(labels.begin(), labels.end(), 0);
		std::random_shuffle(labels.begin(), labels.end());

		int task_count = static_cast<int>(labels.size() / 2);

		#pragma omp parallel for schedule(dynamic)
		for (int i = 0; i < task_count; ++i) {
			assert(static_cast<size_t>(2*i) < labels.size());
			assert(static_cast<size_t>(2*i+1) < labels.size());

			class_t c1 = labels[2*i];
			class_t c2 = labels[2*i+1];
			relabel_c1_c2(c1, c2);
		}
	}

	// Relabel subsets in batches
	void relabel_sweep_subset(size_t subset_size, size_t passes_per_batch = 1) {
		std::vector<size_t> sites(n);
		std::iota(sites.begin(), sites.end(), 0);
		std::random_shuffle(sites.begin(), sites.end());

		for (size_t start = 0; start < n; start += subset_size) {
			size_t end = std::min(n, start + subset_size);
			std::set<size_t> W(sites.begin() + start, sites.begin() + end);

			// Random permutation of labels
			std::vector<class_t> labels(class_count, 0);
			std::iota(labels.begin(), labels.end(), 0);
			for (size_t pass = 0; pass < passes_per_batch; ++pass) {
				std::random_shuffle(labels.begin(), labels.end());

				int task_count = labels.size() / 2;
				#pragma omp parallel for schedule(dynamic)
				for (int i = 0; i < task_count; ++i) {
					assert(static_cast<size_t>(2*i) < labels.size());
					assert(static_cast<size_t>(2*i+1) < labels.size());

					class_t c1 = labels[2*i];
					class_t c2 = labels[2*i+1];
					relabel_c1_c2(W, c1, c2);
				}
			}
		}
	}

	// Relabel subsets, but assure that all possible pairs of labels are
	// processed for each subset.
	void relabel_sweep_subset_pp(size_t subset_size) {
		std::vector<size_t> sites(n);
		std::iota(sites.begin(), sites.end(), 0);
		std::random_shuffle(sites.begin(), sites.end());

		// Partition label sets into independent all-pairs
		pairpartition labelpart(class_count);
		size_t labelpart_count = labelpart.number_of_sets();

		for (size_t start = 0; start < n; start += subset_size) {
			size_t end = std::min(n, start + subset_size);
			std::set<size_t> W(sites.begin() + start, sites.begin() + end);

			for (size_t si = 0; si < labelpart_count; ++si) {
				auto const& set = labelpart.get_set(si);
				int task_count = set.size();

				#pragma omp parallel for
				for (int i = 0; i < task_count; ++i) {
					class_t c1 = set[i].first;
					class_t c2 = set[i].second;
					relabel_c1_c2(W, c1, c2);
				}
			}
		}
	}

	// Initialize from previous solution.
	// Preconditions: y_init.size() == n, y_init[i] < class_count.
	void initialize(const std::vector<class_t>& y_init) {
		assert(y_init.size() == n);
		for (size_t i = 0; i < n; ++i) {
			assert(y_init[i] < class_count);
			y[i] = y_init[i];
		}
	}
	void initialize(initfunc_t y_init) {
		for (size_t i = 0; i < n; ++i)
			y[i] = y_init(i);
	}

	// Initialize solution by either random or MAP assignment.
	void initialize(init_type init, unsigned long seed = 0) {
		assert(y.size() == n);

		if (init == map_assignment) {
			// Find the per-variable MAP
			for (size_t i = 0; i < n; ++i) {
				TFloat max_val = -std::numeric_limits<TFloat>::infinity();
				size_t max_c = 0;
				for (size_t c = 0; c < class_count; ++c) {
					if (P(i,c) <= max_val)
						continue;
					max_val = P(i,c);
					max_c = c;
				}
				y[i] = static_cast<class_t>(max_c);
			}
		} else if (init == random_assignment) {
			// Randomly assign classes
			std::random_device gseed;
			std::mt19937 eng(seed == 0 ? gseed() : seed);
			std::uniform_int_distribution<size_t> rlabel(0, class_count-1);

			for (size_t i = 0; i < n; ++i)
				y[i] = rlabel(eng);
		} else if (init == inverseweighted_assignment) {
			// Compute class frequencies
			std::vector<double> freq(class_count, 0.0);
			for (size_t i = 0; i < n; ++i)
				for (size_t c = 0; c < class_count; ++c)
					freq[c] += P(i,c);

			double sum = std::accumulate(freq.cbegin(), freq.cend(), 0.0);
			for (size_t c = 0; c < class_count; ++c)
				freq[c] /= sum;

			// Find the per-variable reweighted MAP
			for (size_t i = 0; i < n; ++i) {
				TFloat max_val = -std::numeric_limits<TFloat>::infinity();
				size_t max_c = 0;
				for (size_t c = 0; c < class_count; ++c) {
					double P_c_weighted = P(i,c) / freq[c];
					if (P_c_weighted < max_val)
						continue;

					max_val = P_c_weighted;
					max_c = c;
				}
				y[i] = static_cast<class_t>(max_c);
			}
		} else {
			assert(0);
		}
	}
};

}

#endif

