
#include <iostream>
#include <fstream>
#include <random>
#include <cmath>
#include <cassert>

#include "voc.h"
#include "dirichlet.h"
#include "timing.h"

void produce_random_instance(std::vector<float>& P,
	size_t n, size_t class_count, unsigned long seed = 0)
{
	std::random_device gseed;
	std::mt19937 eng(seed == 0 ? gseed() : seed);
	std::uniform_real_distribution<float> runi;

	dirichlet<double> dir(class_count, 0.1);
	std::vector<double> prior;

	P.resize(n*class_count);
	for (size_t i = 0; i < n; ++i) {
		dir.simulate(prior, eng);
		for (size_t c = 0; c < class_count; ++c)
			P[i*class_count + c] = prior[c];
	}
}

void test_greedy(size_t n, size_t class_count, unsigned long seed = 0) {
	std::vector<float> P(n*class_count);
	produce_random_instance(P, n, class_count, seed);

	voc::prediction_function<float> pred(&P[0], class_count, n);
	pred.initialize(voc::prediction_function<float>::random_assignment);

	double avg_sweep_time = 0.0;
	size_t iter;
	for (iter = 1; iter <= 50; ++iter) {
		double obj = pred.compute_objective();
		std::cout << "[voc] GREEDY  iter " << iter << "  score " << obj << std::endl;

		timer t;
		t.start();
		size_t changed = pred.greedy_sweep();
		t.stop();
		avg_sweep_time += t.elapsed_s();
		std::cout << "   greedy sweep took " << t.elapsed_s() << "s, changed "
			<< changed << " variables" << std::endl;
	}
	double final_obj = pred.compute_objective();
	std::cout << "Final score " << final_obj << std::endl;
	std::cout << "Average time per sweep "
		<< (avg_sweep_time / static_cast<double>(iter)) << "s" << std::endl;
}

void test_random(size_t n, size_t class_count, unsigned long seed = 0) {
	std::vector<float> P(n*class_count);
	produce_random_instance(P, n, class_count, seed);

	voc::prediction_function<float> pred(&P[0], class_count, n);
	//pred.initialize(voc::prediction_function<float>::random_assignment);
	//pred.initialize(voc::prediction_function<float>::inverseweighted_assignment);
	pred.initialize(voc::prediction_function<float>::map_assignment);

	double obj_prev = -std::numeric_limits<double>::infinity();
	double avg_sweep_time = 0.0;
	size_t iter;
	for (iter = 1; iter <= 50; ++iter) {
		double obj = pred.compute_objective();
		std::cout << "[voc] iter " << iter << "  score " << obj << std::endl;

		assert(obj >= obj_prev);
		obj_prev = obj;

		timer t;
		t.start();
		pred.relabel_sweep_subset(2000);
		//pred.relabel_sweep();
		t.stop();
		avg_sweep_time += t.elapsed_s();
		//std::cout << "   voc sweep took " << t.elapsed_s() << "s" << std::endl;
	}
	double final_obj = pred.compute_objective();
	std::cout << "Final score " << final_obj << std::endl;
	std::cout << "Average time per sweep "
		<< (avg_sweep_time / static_cast<double>(iter)) << "s" << std::endl;
}

int main(int argc, char* argv[]) {
	test_greedy(10000, 21);
	test_random(10000, 21);

#if 0
	for (size_t stress = 0; stress < 10000; ++stress)
		test_random(2, 2, stress);
#endif
}


